"""Tempfile utils."""

import tempfile
from pathlib import Path


class TemporaryDirectory:
    """Wrapper over temporary directory to output paths."""

    def __init__(self) -> None:
        self.temporary_directory = (
            tempfile.TemporaryDirectory()  # pylint: disable=consider-using-with
        )

    def __enter__(self) -> Path:
        return Path(self.temporary_directory.__enter__())

    def __exit__(self, type_, value, traceback) -> None:
        self.temporary_directory.__exit__(type_, value, traceback)



"""GraphQL payload data mappers for api keys operations."""

from kili.domain.api_key import ApiKeyFilters


def api_key_where_mapper(filters: ApiKeyFilters):
    """Build the GraphQL ApiKeyWhere variable to be sent in an operation."""
    return {
        "user": {"id": filters.user_id},
        "id": filters.api_key_id,
        "key": filters.api_key,
    }



"""Helpers for the issue mutations."""

from typing import Dict

from kili.adapters.kili_api_gateway.helpers.queries import QueryOptions
from kili.adapters.kili_api_gateway.kili_api_gateway import KiliAPIGateway
from kili.domain.label import LabelFilters, LabelId
from kili.domain.project import ProjectId
from kili.domain.types import ListOrTuple
from kili.exceptions import NotFound


def get_labels_asset_ids_map(
    kili_api_gateway: KiliAPIGateway,
    project_id: ProjectId,
    label_id_array: ListOrTuple[LabelId],
) -> Dict:
    """Return a dictionary that gives for every label id, its associated asset id.

    Args:
        kili_api_gateway: instance of KiliAPIGateway
        project_id: id of the project
        label_id_array: list of label ids

    Returns:
        a dict of key->value a label id->its associated asset id for the given label ids

    Raises:
        NotFound error if at least one label was not found with its given id
    """
    options = QueryOptions(disable_tqdm=True)
    labels = list(
        kili_api_gateway.list_labels(
            filters=LabelFilters(project_id=project_id, id_in=label_id_array),
            fields=("labelOf.id", "id"),
            options=options,
        )
    )

    labels_not_found = [
        label_id for label_id in label_id_array if label_id not in [label["id"] for label in labels]
    ]
    if len(labels_not_found) > 0:
        raise NotFound(str(labels_not_found))
    return {label["id"]: label["labelOf"]["id"] for label in labels}



"""GraphQL payload data mappers for user operations."""

from typing import Dict

from kili.domain.user import UserFilter

from .types import CreateUserDataKiliGatewayInput, UserDataKiliGatewayInput


def user_where_mapper(filters: UserFilter) -> Dict:
    """Build the GraphQL UserWhere variable to be sent in an operation."""
    return {
        "activated": filters.activated,
        "apiKey": filters.api_key,
        "email": filters.email,
        "id": filters.id,
        "idIn": filters.id_in,
        "organization": {"id": filters.organization_id},
    }


def create_user_data_mapper(data: CreateUserDataKiliGatewayInput) -> Dict:
    """Build the CreateUserDataKiliGatewayInput data variable to be sent in an operation."""
    return {
        "email": data.email.lower(),
        "firstname": data.firstname,
        "lastname": data.lastname,
        "password": data.password,
        "organizationRole": data.organization_role,
    }


def update_user_data_mapper(data: UserDataKiliGatewayInput) -> Dict:
    """Build the UserDataKiliGatewayInput data variable to be sent in an operation."""
    return {
        "activated": data.activated,
        "apiKey": data.api_key,
        # "auth0Id": data.auth0_id,  # refused by the backend: only used for service account  # noqa: ERA001  # pylint: disable=line-too-long
        "email": data.email,
        "firstname": data.firstname,
        "hasCompletedLabelingTour": data.has_completed_labeling_tour,
        "hubspotSubscriptionStatus": data.hubspot_subscription_status,
        "lastname": data.lastname,
        "organization": data.organization,
        "organizationId": data.organization_id,
        "organizationRole": data.organization_role,
    }



"""Authentification functions that does not rely on the graphQL client."""

from kili import __version__
from kili.adapters.http_client import HttpClient
from kili.core.graphql.clientnames import GraphQLClientName
from kili.utils.logcontext import LogContext, log_call


@log_call
def is_api_key_valid(
    http_client: HttpClient,
    api_key: str,
    api_endpoint: str,
    client_name: GraphQLClientName,
) -> bool:
    """Check that the api_key provided is valid.

    Note that this method does not rely on the GraphQL client, but on the HTTP client.
    It must stay this way since the GraphQL client might retry in case of 401 http error.
    """
    response = http_client.post(
        url=api_endpoint,
        data='{"query":"{ me { id email } }"}',
        timeout=30,
        headers={
            "Authorization": f"X-API-Key: {api_key}",
            "Accept": "application/json",
            "Content-Type": "application/json",
            "apollographql-client-name": client_name.value,
            "apollographql-client-version": __version__,
            **LogContext(),
        },
    )
    return (
        response.status_code == 200  # noqa: PLR2004
        and "email" in response.text
        and "id" in response.text
    )



"""Issue mutations."""

from typing import Dict, Optional

from typeguard import typechecked

from kili.adapters.kili_api_gateway.helpers.queries import QueryOptions
from kili.adapters.kili_api_gateway.issue.operations import GQL_CREATE_ISSUES
from kili.core.helpers import deprecate
from kili.domain.issue import IssueType
from kili.domain.label import LabelFilters, LabelId
from kili.domain.project import ProjectId
from kili.entrypoints.base import BaseOperationEntrypointMixin
from kili.use_cases.label import LabelUseCases
from kili.utils.logcontext import for_all_methods, log_call


@for_all_methods(log_call, exclude=["__init__"])
class MutationsIssue(BaseOperationEntrypointMixin):
    """Set of Issue mutations."""

    # pylint: disable=too-many-arguments
    @deprecate(
        msg=(
            "append_to_issues is deprecated. Please use `create_issues` or `create_questions`"
            " instead. These new methods allow to import several issues or questions at the same"
            " time and provide better performances."
        )
    )
    @typechecked
    def append_to_issues(
        self,
        label_id: str,
        project_id: str,
        object_mid: Optional[str] = None,
        text: Optional[str] = None,
        type_: IssueType = "ISSUE",
    ) -> Dict:
        """Create an issue.

        !!! danger "Deprecated"
            append_to_issues is deprecated.
            Please use `create_issues` or `create_questions` instead.
            These new methods allow to import several issues or questions at the same time
            and provide better performances.

        Args:
            label_id: Id of the label to add an issue to
            object_mid: Mid of the object in the label to associate the issue to
            type_: type of the issue to add. Can be either "ISSUE" or "QUESTION"
            text: If given, write a comment related to the issue
            project_id: Id of the project

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.
        """
        labels = list(
            LabelUseCases(self.kili_api_gateway).list_labels(
                filters=LabelFilters(id=LabelId(label_id), project_id=ProjectId(project_id)),
                project_id=ProjectId(project_id),
                fields=("assetId",),
                output_format="dict",
                options=QueryOptions(disable_tqdm=True, first=1),
            )
        )
        if not labels:
            raise ValueError(
                f"Label ID {label_id} does not exist in the project of ID {project_id}"
            )

        asset_id = labels[0]["assetId"]
        variables = {
            "issues": [
                {
                    "issueNumber": 0,
                    "labelID": label_id,
                    "objectMid": object_mid,
                    "type": type_,
                    "assetId": asset_id,
                    "text": text,
                }
            ],
            "where": {"id": asset_id},
        }

        result = self.graphql_client.execute(GQL_CREATE_ISSUES, variables)
        return self.format_result("data", result)[0]



"""Types for the Coco export."""

from typing import Dict, List

from typing_extensions import TypedDict


class CocoImage(TypedDict):
    """Handle the coco image data."""

    id: int
    license: int
    file_name: str
    height: int
    width: int
    date_captured: None


class CocoCategory(TypedDict):
    """Handle the coco category data."""

    id: int
    name: str
    supercategory: str


class CocoAnnotation(TypedDict):
    """Handle the coco annotation data."""

    id: int
    image_id: int
    category_id: int
    bbox: List[int]
    segmentation: List[List[float]]  # [[x, y, x, y, x ...]]
    area: int
    iscrowd: int


class CocoFormat(TypedDict):
    """Handle the coco format data."""

    info: Dict  # type: ignore
    licenses: List[Dict]  # type: ignore
    categories: List[CocoCategory]
    images: List[CocoImage]
    annotations: List[CocoAnnotation]



"""Functions to activate/deactivate a plugin."""

from typing import Optional

from kili.core.graphql.operations.plugin.mutations import (
    GQL_ACTIVATE_PLUGIN_ON_PROJECT,
    GQL_DEACTIVATE_PLUGIN_ON_PROJECT,
)
from kili.services.plugins.helpers import get_logger
from kili.services.plugins.tools import check_errors_plugin_activation


def activate_plugin(kili, plugin_name: str, project_id: str) -> Optional[str]:
    """Create a plugin in Kili."""
    logger = get_logger()

    variables = {"pluginName": plugin_name, "projectId": project_id}

    result = kili.graphql_client.execute(GQL_ACTIVATE_PLUGIN_ON_PROJECT, variables)

    has_failed, already_activated = check_errors_plugin_activation(result, plugin_name, project_id)

    if not already_activated and not has_failed:
        logger.info(f'Plugin with name "{plugin_name}" activated on project "{project_id}"')

    return kili.format_result("data", result) if not already_activated else None


def deactivate_plugin(kili, plugin_name: str, project_id: str) -> str:
    """Create a plugin in Kili."""
    logger = get_logger()

    variables = {"pluginName": plugin_name, "projectId": project_id}

    result = kili.graphql_client.execute(GQL_DEACTIVATE_PLUGIN_ON_PROJECT, variables)

    logger.info(f"Plugin {plugin_name} deactivated on project {project_id}")

    return kili.format_result("data", result)



"""Label use cases."""

from functools import partial
from typing import TYPE_CHECKING, Dict, Generator, List, Literal, Optional

from kili.adapters.kili_api_gateway.helpers.queries import QueryOptions
from kili.adapters.kili_api_gateway.label.types import (
    AppendLabelData,
    AppendManyLabelsData,
    AppendToLabelsData,
    UpdateLabelData,
)
from kili.domain.asset import AssetExternalId, AssetFilters, AssetId
from kili.domain.label import LabelFilters, LabelId, LabelType
from kili.domain.project import ProjectId
from kili.domain.types import ListOrTuple
from kili.domain.user import UserId
from kili.use_cases.asset.utils import AssetUseCasesUtils
from kili.use_cases.base import BaseUseCases
from kili.utils.labels.parsing import parse_labels

from .types import LabelToCreateUseCaseInput
from .validator import check_input_labels

if TYPE_CHECKING:
    import pandas as pd


class LabelUseCases(BaseUseCases):
    """Label use cases."""

    def count_labels(self, filters: LabelFilters) -> int:
        """Count labels."""
        return self._kili_api_gateway.count_labels(filters=filters)

    def list_labels(
        self,
        project_id: ProjectId,
        filters: LabelFilters,
        fields: ListOrTuple[str],
        options: QueryOptions,
        output_format: Literal["dict", "parsed_label"],
    ) -> Generator:
        """List labels."""
        label_parser_post_function = None
        if output_format == "parsed_label":
            if "jsonResponse" not in fields:
                raise ValueError(
                    "The field 'jsonResponse' is required to parse labels. Please add it to the"
                    " 'fields' argument."
                )

            project = self._kili_api_gateway.get_project(
                project_id, fields=("jsonInterface", "inputType")
            )

            label_parser_post_function = partial(
                parse_labels,
                json_interface=project["jsonInterface"],
                input_type=project["inputType"],
            )

        labels_gen = self._kili_api_gateway.list_labels(
            fields=fields, filters=filters, options=options
        )

        if label_parser_post_function is not None:
            labels_gen = label_parser_post_function(labels=labels_gen)

        return labels_gen

    def update_properties_in_label(
        self,
        label_id: LabelId,
        seconds_to_label: Optional[int],
        model_name: Optional[str],
        json_response: Optional[Dict],
        fields: ListOrTuple[str],
    ) -> Dict:
        """Update properties in label."""
        return self._kili_api_gateway.update_properties_in_label(
            label_id=label_id,
            data=UpdateLabelData(
                json_response=json_response,
                model_name=model_name,
                seconds_to_label=seconds_to_label,
                is_sent_back_to_queue=None,
            ),
            fields=fields,
        )

    def delete_labels(
        self, ids: ListOrTuple[LabelId], disable_tqdm: Optional[bool]
    ) -> List[LabelId]:
        """Delete labels."""
        return self._kili_api_gateway.delete_labels(ids=ids, disable_tqdm=disable_tqdm)

    def append_labels(
        self,
        labels: List[LabelToCreateUseCaseInput],
        label_type: LabelType,
        overwrite: Optional[bool],
        project_id: Optional[ProjectId],
        fields: ListOrTuple[str],
        disable_tqdm: Optional[bool],
    ) -> List[Dict]:
        """Append labels."""
        check_input_labels(labels)

        asset_id_array = [label.asset_id for label in labels]
        if any(asset_id is None for asset_id in asset_id_array):
            external_id_array = [label.asset_external_id for label in labels]
            asset_id_array = AssetUseCasesUtils(
                self._kili_api_gateway
            ).get_asset_ids_or_throw_error(
                asset_ids=None,
                external_ids=external_id_array,  # pyright: ignore[reportGeneralTypeIssues]
                project_id=project_id,
            )

        labels_to_add = [
            AppendLabelData(
                author_id=label.author_id,
                asset_id=asset_id,  # pyright: ignore[reportGeneralTypeIssues]
                seconds_to_label=label.seconds_to_label,
                json_response=label.json_response,
                model_name=label.model_name,
                client_version=None,
            )
            for label, asset_id in zip(labels, asset_id_array)
        ]

        data = AppendManyLabelsData(
            label_type=label_type,
            overwrite=overwrite,
            labels_data=labels_to_add,
        )
        return self._kili_api_gateway.append_many_labels(
            fields=fields,
            disable_tqdm=disable_tqdm,
            data=data,
            project_id=project_id,
        )

    def append_to_labels(
        self,
        author_id: Optional[UserId],
        json_response: Dict,
        label_type: LabelType,
        seconds_to_label: Optional[float],
        asset_id: AssetId,
        fields: ListOrTuple[str],
    ) -> Dict:
        """Append to labels."""
        data = AppendToLabelsData(
            author_id=(
                author_id
                if author_id is not None
                else self._kili_api_gateway.get_current_user(fields=("id",))["id"]
            ),
            json_response=json_response,
            label_type=label_type,
            seconds_to_label=seconds_to_label,
            client_version=None,
            skipped=None,
        )
        return self._kili_api_gateway.append_to_labels(data=data, asset_id=asset_id, fields=fields)

    def create_honeypot_label(
        self,
        json_response: Dict,
        asset_id: Optional[AssetId],
        asset_external_id: Optional[AssetExternalId],
        project_id: Optional[ProjectId],
        fields: ListOrTuple[str],
    ) -> Dict:
        """Create honeypot label."""
        if asset_id is None:
            if asset_external_id is None or project_id is None:
                raise ValueError(
                    "Either provide `asset_id` or `asset_external_id` and `project_id`."
                )

            asset_id = AssetUseCasesUtils(self._kili_api_gateway).infer_ids_from_external_ids(
                asset_external_ids=[asset_external_id],
                project_id=project_id,
            )[asset_external_id]

        return self._kili_api_gateway.create_honeypot_label(
            json_response=json_response, asset_id=asset_id, fields=fields
        )

    def export_labels_as_df(
        self,
        *,
        project_id: ProjectId,
        label_fields: ListOrTuple[str],
        asset_fields: ListOrTuple[str],
    ) -> "pd.DataFrame":
        """Export labels as a pandas DataFrame."""
        assets_gen = self._kili_api_gateway.list_assets(
            AssetFilters(project_id=ProjectId(project_id)),
            tuple(asset_fields) + tuple("labels." + field for field in label_fields),
            QueryOptions(disable_tqdm=False),
        )

        labels = [
            dict(
                label,
                **{f"asset_{key}": asset[key] for key in asset if key != "labels"},
            )
            for asset in assets_gen
            for label in asset["labels"]
        ]

        import pandas as pd  # pylint: disable=import-outside-toplevel

        return pd.DataFrame(labels)



"""Classes for json response parsing."""

from typing import Dict, Iterator, List, Optional, Tuple, cast

from kili.services.label_data_parsing import job_response as job_response_module

from .exceptions import FrameIndexError, JobNotExistingError
from .types import Project

IGNORED_JOBS = ["ANNOTATION_JOB_COUNTER", "ANNOTATION_NAMES_JOB", "ROTATION_JOB"]


# pylint: disable=too-few-public-methods
class _ParsedVideoJobs:
    def __init__(
        self,
        project_info: Project,
        json_response: Dict,
        job_names_to_parse: Optional[List[str]] = None,
    ) -> None:
        self._json_data: Dict[str, "FramesList"] = {}

        self._nb_frames = len(json_response)

        json_interface = project_info["jsonInterface"]

        # all job names in the json response should be in the json interface too
        for frame_response in json_response.values():
            for job_name in frame_response:
                if job_name not in json_interface and job_name not in IGNORED_JOBS:
                    raise JobNotExistingError(job_name)

        # define the list of job names to parse
        # by default, parse all the jobs that are not children
        if job_names_to_parse is None:
            job_names_to_parse = [
                job_name
                for job_name, job_interface in json_interface.items()
                if not job_interface["isChild"]
            ]

        for current_job_name in job_names_to_parse:
            frames_list_for_job = []
            for _, frame_json_response in sorted(
                json_response.items(),
                key=lambda item: int(item[0]),  # sort by frame number
            ):
                job_response = frame_json_response.get(current_job_name, {})
                job_payload = job_response_module.JobPayload(
                    job_name=current_job_name,
                    project_info=project_info,
                    job_payload=job_response,
                )
                frames_list_for_job.append(job_payload)

            assert len(frames_list_for_job) == self._nb_frames, (
                f"len(frames_list_for_job) = {len(frames_list_for_job)} != {self._nb_frames} ="
                " self._nb_frames"
            )
            self._json_data[current_job_name] = FramesList(frames_list_for_job)

    def to_dict(self) -> Dict[str, Dict]:
        """Returns a copy of the parsed label as a dict."""
        ret = {str(frame_id): {} for frame_id in range(self._nb_frames)}
        for job_name, frames_list in self._json_data.items():
            for frame_id, job_payload in enumerate(frames_list):
                job_payload_dict = job_payload.to_dict()
                if job_payload_dict:
                    ret[str(frame_id)][job_name] = job_payload_dict
        return ret


# pylint: disable=too-few-public-methods
class _ParsedJobs:
    def __init__(
        self,
        project_info: Project,
        json_response: Dict,
        job_names_to_parse: Optional[List[str]] = None,
    ) -> None:
        self._json_data: Dict[str, "job_response_module.JobPayload"] = {}

        json_interface = project_info["jsonInterface"]

        # all job names in the json response should be in the json interface too
        for job_name in json_response:
            if job_name not in json_interface:
                raise JobNotExistingError(job_name)

        # define the list of job names to parse
        # by default, parse all the jobs that are not children
        if job_names_to_parse is None:
            job_names_to_parse = [
                job_name
                for job_name, job_interface in json_interface.items()
                if not job_interface["isChild"]
            ]

        for job_name in job_names_to_parse:
            job_response = json_response.get(job_name, {})  # the json response can be empty
            self._json_data[job_name] = job_response_module.JobPayload(
                job_name=job_name,
                project_info=project_info,
                job_payload=job_response,
            )

    def to_dict(self) -> Dict[str, Dict]:
        """Returns the parsed json response as a dict."""
        ret = {job_name: job_payload.to_dict() for job_name, job_payload in self._json_data.items()}
        return {k: v for k, v in ret.items() if v}  # remove empty json responses


def _is_video_response(project_info: Project, json_response: Dict) -> bool:
    """Returns True if the json response is a video job, False otherwise."""
    if "VIDEO" not in project_info["inputType"]:
        return False

    if not all(frame_id.isdigit() for frame_id in json_response):
        return False

    for frame_response in json_response.values():
        if len(frame_response) == 0:
            continue

        for job_name in frame_response:
            if (
                isinstance(job_name, str)
                and job_name not in IGNORED_JOBS
                and job_name not in project_info["jsonInterface"]
            ):
                return False

    return True


class ParsedJobs(_ParsedJobs, _ParsedVideoJobs):
    """Class for label json response parsing."""

    def __init__(
        self,
        project_info: Project,
        json_response: Dict,
        job_names_to_parse: Optional[List[str]] = None,
    ) -> None:
        # pylint: disable=line-too-long
        """Class for label json response parsing.

        This class will modify the input json_response.
        If you want to keep the original json_response, use deepcopy.

        Args:
            project_info: Information about the project.
            json_response: Value of the key "jsonResponse" of a label.
            job_names_to_parse: List of job names to parse. By default, parse all the jobs that are not children.
        """
        self._is_video_response = _is_video_response(project_info, json_response)

        if self._is_video_response:
            _ParsedVideoJobs.__init__(
                self,
                project_info=project_info,
                json_response=json_response,
                job_names_to_parse=job_names_to_parse,
            )
        else:
            _ParsedJobs.__init__(
                self,
                project_info=project_info,
                json_response=json_response,
                job_names_to_parse=job_names_to_parse,
            )

    def to_dict(self) -> Dict[str, Dict]:
        """Returns the parsed json response as a dict."""
        if self._is_video_response:
            return _ParsedVideoJobs.to_dict(self)
        return _ParsedJobs.to_dict(self)

    def __repr__(self) -> str:
        """Returns the representation of the object."""
        return repr(self.to_dict())

    def __str__(self) -> str:
        """Returns the string representation of the object."""
        return str(self.to_dict())

    def __len__(self) -> int:
        """Returns the number of jobs."""
        return len(self._json_data)

    def __iter__(self) -> Iterator[str]:
        """Returns an iterator over the job names."""
        return iter(self._json_data)

    def items(self) -> Iterator[Tuple[str, "JobPayload"]]:
        """Returns an iterator over the job names and the corresponding objects."""
        return cast(Iterator[Tuple[str, "JobPayload"]], iter(self._json_data.items()))

    def keys(self) -> Iterator[str]:
        """Returns an iterator over the job names."""
        return iter(self._json_data.keys())

    def values(self) -> Iterator["JobPayload"]:
        """Returns an iterator over the objects."""
        return cast(Iterator["JobPayload"], iter(self._json_data.values()))

    def __getitem__(self, job_name: str) -> "JobPayload":
        """Returns the object corresponding to the job name.

        For a video job, the returned object is a FramesList.

        Otherwise, the returned object is a JobPayload.
        """
        if job_name not in self._json_data:
            raise JobNotExistingError(job_name)
        return cast("JobPayload", self._json_data[job_name])


class FramesList(List["job_response_module.JobPayload"]):
    """List class that allows to access the JobPayload object corresponding to a frame number."""

    def __getitem__(self, key: int) -> "job_response_module.JobPayload":
        """Returns the JobPayload object corresponding to the frame number."""
        if not 0 <= key <= len(self) - 1:
            raise FrameIndexError(frame_index=key, nb_frames=len(self))
        return super().__getitem__(key)

    @property
    def frames(self) -> "FramesList":
        """Returns the list of frames."""
        return self


class JobPayload(FramesList, job_response_module.JobPayload):
    # pylint: disable=line-too-long
    """JobPayload class.

    It is the value of the parsed job response of a job `job_name`: `label.jobs["job_name"]`.

    If the job is a video job, it can be used to access the job response of a frame `frame_number`: `label.jobs["job_name"].frames[frame_number]`.
    """

    # Class only used for return type hints.
    # It should not be instantiated.
    # It inherits from both FramesList and JobResponseModule.JobPayload to expose the methods
    # and properties of both classes.



"""LLM domain."""

from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict, Optional, Union

from typing_extensions import TypedDict


@dataclass
class OrganizationModelFilters:
    """Model filters for running a model search."""

    organization_id: str


class ModelType(str, Enum):
    """Enumeration of the supported model types.

    - `AZURE_OPEN_AI`: Models hosted on Microsoft Azure's OpenAI service.
    - `OPEN_AI_SDK`: Models provided via OpenAI's SDK.
    """

    AZURE_OPEN_AI = "AZURE_OPEN_AI"
    OPEN_AI_SDK = "OPEN_AI_SDK"


@dataclass
class AzureOpenAICredentials:
    """Credentials for accessing Azure OpenAI models.

    Attributes:
    - `api_key`: The API key required for authentication to Azure OpenAI.
    - `deployment_id`: The specific deployment of the model within Azure.
    - `endpoint`: The endpoint URL where the Azure OpenAI service is hosted.
    """

    api_key: str
    deployment_id: str
    endpoint: str


@dataclass
class OpenAISDKCredentials:
    """Credentials for accessing OpenAI SDK models.

    Attributes:
    - `api_key`: The API key required for authentication to OpenAI's SDK.
    - `endpoint`: The endpoint URL where the OpenAI SDK service is hosted.
    """

    api_key: str
    endpoint: str


@dataclass
class ModelToCreateInput:
    """Model to create use case input."""

    credentials: Union[AzureOpenAICredentials, OpenAISDKCredentials]
    name: str
    type: ModelType
    organization_id: str


@dataclass
class ModelToUpdateInput:
    """Model to update use case input."""

    credentials: Optional[Union[AzureOpenAICredentials, OpenAISDKCredentials]] = None
    name: Optional[str] = None


@dataclass
class ProjectModelToCreateInput:
    """Project model to create use case input."""

    project_id: str
    model_id: str
    configuration: dict


@dataclass
class ProjectModelToUpdateInput:
    """Project model to update use case input."""

    configuration: Optional[dict] = None


@dataclass
class ProjectModelFilters:
    """Project model filters for running a project model search."""

    project_id: Optional[str] = None
    model_id: Optional[str] = None


class ChatItemRole(str, Enum):
    """Enumeration of the supported chat item role."""

    ASSISTANT = "ASSISTANT"
    USER = "USER"


class OpenAISDKCredentialsDict(TypedDict):
    """Dict that represents model.Credentials for OpenAI SDK."""

    api_key: str
    endpoint: str


class AzureOpenAICredentialsDict(TypedDict):
    """Dict that represents model.Credentials for Azure OpenAI."""

    api_key: str
    endpoint: str
    deployment_id: str


class ModelDict(TypedDict):
    """Dict that represents a Model."""

    id: str
    credentials: Union[AzureOpenAICredentialsDict, OpenAISDKCredentialsDict]
    name: str
    type: str


class ProjectModelDict(TypedDict):
    """Dict that represents a ProjectModel."""

    id: str
    configuration: Dict[str, Any]
    model: ModelDict


class ChatItemDict(TypedDict):
    """Dict that represents a ChatItem."""

    content: str
    id: str
    role: ChatItemRole



"""Classes for job response parsing."""

from datetime import datetime
from typing import Dict, List, Optional, cast

from typeguard import typechecked

from kili.services.label_data_parsing import annotation as annotation_module
from kili.services.label_data_parsing import category as category_module
from kili.services.label_data_parsing import json_response as json_response_module
from kili.services.types import Job

from .exceptions import AttributeNotCompatibleWithJobError, InvalidMutationError
from .types import Project
from .utils import get_children_job_names


class JobPayload:
    """Class for job payload parsing."""

    def __init__(self, job_name: str, project_info: Project, job_payload: Dict) -> None:
        """Class for job payload parsing.

        Args:
            job_name: Name of the job.
            project_info: Information about the project.
            job_payload: Value of the key "job_name" of the job response.
        """
        self._job_name = job_name
        self._project_info = project_info
        self._json_data = job_payload

        self._job_interface = project_info["jsonInterface"][job_name]  # type: ignore

        # cast lists to objects
        if "categories" in self._json_data:
            self._json_data["categories"] = category_module.CategoryList(
                job_name=self._job_name,
                project_info=self._project_info,
                categories_list=self._json_data["categories"],
            )

        if "annotations" in self._json_data:
            self._json_data["annotations"] = annotation_module.AnnotationList(
                job_name=self._job_name,
                project_info=self._project_info,
                annotations_list=self._json_data["annotations"],
            )

        if self._json_data.get("children"):
            self.children = self._json_data["children"]

    def to_dict(self) -> Dict:
        """Returns the parsed job payload as a dict."""
        ret = {
            k: v
            for k, v in self._json_data.items()
            if k not in ("categories", "annotations", "children")
        }

        # cast back to python native types
        if "categories" in self._json_data:
            ret["categories"] = (
                self._json_data["categories"]
                if isinstance(self._json_data["categories"], List)
                else self._json_data["categories"].as_list()
            )
        if "annotations" in self._json_data:
            ret["annotations"] = (
                self._json_data["annotations"]
                if isinstance(self._json_data["annotations"], List)
                else self._json_data["annotations"].as_list()
            )
        if "children" in self._json_data:
            ret["children"] = (
                self._json_data["children"].to_dict()
                if not isinstance(self._json_data["children"], Dict)
                else self._json_data["children"]
            )
        return ret

    def __repr__(self) -> str:
        """Returns the parsed job payload representation as a string."""
        return repr(self.to_dict())

    def __str__(self) -> str:
        """Returns the parsed job payload as a string."""
        return str(self.to_dict())

    @property
    def categories(self) -> "category_module.CategoryList":
        """Returns a list of Category objects for a classification job."""
        if self._job_interface["mlTask"] != "CLASSIFICATION":
            raise AttributeNotCompatibleWithJobError("categories")

        if "categories" not in self._json_data and not self._job_interface["required"]:
            return category_module.CategoryList(
                categories_list=[], project_info=self._project_info, job_name=self._job_name
            )

        return self._json_data["categories"]

    @property
    def category(self) -> "category_module.Category":
        """Returns a Category object for a classification job if there is only one category.

        Else raises an error.
        """
        if self._job_interface["mlTask"] != "CLASSIFICATION":
            raise AttributeNotCompatibleWithJobError("category")

        if self._job_interface["content"]["input"] not in ("radio", "singleDropdown"):
            raise AttributeNotCompatibleWithJobError("category")

        if "categories" not in self._json_data and not self._job_interface["required"]:
            return None  # type: ignore

        if len(self._json_data["categories"]) != 1:
            raise ValueError(f"Expected 1 category, got {self._json_data['categories']}")

        return self._json_data["categories"][0]

    @property
    def children(self) -> "json_response_module.ParsedJobs":
        """Returns the parsed children jobs of the job."""
        return self._json_data["children"]

    @children.setter
    @typechecked
    def children(self, children: Dict) -> None:
        """Sets the children jobs of the job."""
        job_names_to_parse = get_children_job_names(
            json_interface=self._project_info["jsonInterface"],
            job_interface=self._job_interface,  # type: ignore
        )
        parsed_children_job = json_response_module.ParsedJobs(
            project_info=self._project_info,
            json_response=children,
            job_names_to_parse=job_names_to_parse,
        )
        self._json_data["children"] = parsed_children_job

    @typechecked
    def add_category(self, name: str, confidence: Optional[float] = None) -> None:
        """Adds a category to a job with categories."""
        if self._job_interface["mlTask"] != "CLASSIFICATION":
            raise AttributeNotCompatibleWithJobError("add_category")

        if "categories" not in self._json_data:
            category_list = category_module.CategoryList(
                categories_list=[], project_info=self._project_info, job_name=self._job_name
            )
            category_list.add_category(name=name, confidence=confidence)
            self._json_data["categories"] = category_list
        else:
            self._json_data["categories"].add_category(name=name, confidence=confidence)

    @property
    def text(self) -> str:
        """Returns the text for a transcription job."""
        if self._job_interface["mlTask"] != "TRANSCRIPTION":
            raise AttributeNotCompatibleWithJobError("text")
        return self._json_data["text"]

    @text.setter
    @typechecked
    def text(self, text: str) -> None:
        """Sets the text for a transcription job."""
        if self._job_interface["mlTask"] != "TRANSCRIPTION":
            raise AttributeNotCompatibleWithJobError("text")

        transcription_field_type = self._job_interface.get("content", {}).get("input", "")

        if transcription_field_type == "date":
            try:
                datetime.strptime(text, r"%Y-%m-%d")
            except ValueError as err:
                raise InvalidMutationError(f"Expected a date, got {text}") from err

        elif transcription_field_type == "number" and not text.isnumeric():
            raise InvalidMutationError(f"Expected a number, got {text}")

        self._json_data["text"] = text

    @property
    def is_key_frame(self) -> bool:
        """Returns the value of isKeyFrame for a video job.

        This is a Boolean indicating if the timestamp or frame is used for interpolation.
        """
        if self._job_interface["mlTask"] != "CLASSIFICATION":
            raise AttributeNotCompatibleWithJobError("is_key_frame")
        return self._json_data["isKeyFrame"]

    @property
    def annotations(self) -> "annotation_module.AnnotationList":
        """Returns a list of Annotation objects for a job."""
        if not _can_query_annotations(json_data=self._json_data, job_interface=self._job_interface):
            raise AttributeNotCompatibleWithJobError("annotations")

        if "annotations" not in self._json_data and not self._job_interface["required"]:
            return annotation_module.AnnotationList(
                annotations_list=[], project_info=self._project_info, job_name=self._job_name
            )

        return self._json_data["annotations"]

    @property
    def entity_annotations(self) -> List["annotation_module.EntityAnnotation"]:
        """Returns a list of EntityAnnotation objects for a named entities recognition job."""
        if self._job_interface["mlTask"] != "NAMED_ENTITIES_RECOGNITION":
            raise AttributeNotCompatibleWithJobError("entity_annotations")

        if not _can_query_annotations(json_data=self._json_data, job_interface=self._job_interface):
            raise AttributeNotCompatibleWithJobError("entity_annotations")

        return cast(List["annotation_module.EntityAnnotation"], self.annotations)

    @property
    def bounding_poly_annotations(self) -> List["annotation_module.BoundingPolyAnnotation"]:
        """Returns a list of BoundingPolyAnnotation objects for an object detection job."""
        if self._job_interface["mlTask"] != "OBJECT_DETECTION":
            raise AttributeNotCompatibleWithJobError("bounding_poly_annotations")

        if not _can_query_annotations(json_data=self._json_data, job_interface=self._job_interface):
            raise AttributeNotCompatibleWithJobError("bounding_poly_annotations")

        return cast(List["annotation_module.BoundingPolyAnnotation"], self.annotations)

    @typechecked
    def add_annotation(self, annotation_dict: Dict) -> None:
        """Adds an annotation to a job with annotations."""
        if not _can_query_annotations(json_data=self._json_data, job_interface=self._job_interface):
            raise AttributeNotCompatibleWithJobError("add_annotation")

        if "annotations" not in self._json_data:
            annotation_list = annotation_module.AnnotationList(
                annotations_list=[], project_info=self._project_info, job_name=self._job_name
            )
            annotation_list.add_annotation(annotation_dict)
            self._json_data["annotations"] = annotation_list
        else:
            self._json_data["annotations"].add_annotation(annotation_dict)


def _can_query_annotations(json_data: Dict, job_interface: Job) -> bool:
    """Checks if the "annotations" key can be queried for the job."""
    if "annotations" in json_data:
        return True

    if job_interface["mlTask"] in ("CLASSIFICATION",):
        return False

    if not job_interface["required"]:
        return True

    return True



"""Label import service."""

from pathlib import Path
from typing import TYPE_CHECKING, List, Optional, Type, cast

from kili.domain.project import ProjectId
from kili.exceptions import NotFound
from kili.services.helpers import is_target_job_in_json_interface
from kili.services.label_import.importer import (
    AbstractLabelImporter,
    KiliRawLabelImporter,
    LoggerParams,
    YoloLabelImporter,
)
from kili.services.label_import.types import LabelFormat
from kili.services.types import LogLevel

if TYPE_CHECKING:
    from kili.client import Kili


def import_labels_from_files(  # pylint: disable=too-many-arguments
    kili: "Kili",
    labels_files: List[str],
    meta_file_path: Optional[str],
    project_id: ProjectId,
    input_format: str,
    target_job_name: Optional[str],
    disable_tqdm: Optional[bool],
    log_level: str,
    model_name: Optional[str],
    is_prediction: bool,
    overwrite: bool,
) -> None:
    """Imports labels from a list of files contained in file path."""
    kili.kili_api_gateway.get_project(project_id, ["id"])

    if len(labels_files) == 0:
        raise ValueError("You must specify files to upload")

    if is_prediction and model_name is None:
        raise ValueError("If predictions are uploaded, a model name should be specified")

    if target_job_name and not is_target_job_in_json_interface(kili, project_id, target_job_name):
        raise NotFound(
            f"Target job {target_job_name} has not been found in the project JSON interface"
        )

    label_importer_class: Optional[Type[AbstractLabelImporter]] = None
    if input_format in ["yolo_v4", "yolo_v5", "yolo_v7"]:
        label_importer_class = YoloLabelImporter
    elif input_format in ["raw", "kili"]:
        label_importer_class = KiliRawLabelImporter
    else:
        raise NotImplementedError(f"{input_format} import is not implemented yet.")

    logger_params = LoggerParams(disable_tqdm=disable_tqdm, level=cast(LogLevel, log_level))
    label_importer = label_importer_class(kili, logger_params, cast(LabelFormat, input_format))
    label_importer.process_from_files(
        [Path(lf) for lf in labels_files],
        Path(meta_file_path) if meta_file_path is not None else None,
        cast(ProjectId, project_id),
        target_job_name,
        model_name,
        is_prediction,
        overwrite,
    )



"""Service exceptions."""


class NotEnoughArgumentsSpecifiedError(ValueError):
    """Raised when there are not enough arguments specified in a service."""


class TooManyArgumentsSpecifiedError(ValueError):
    """Raised when there are too many arguments specified in a service."""



"""Label domain."""

from dataclasses import dataclass
from typing import TYPE_CHECKING, Literal, NewType, Optional

from kili.domain.types import ListOrTuple

if TYPE_CHECKING:
    from .asset import AssetFilters
    from .project import ProjectId
    from .user import UserFilter

LabelId = NewType("LabelId", str)


LabelType = Literal["AUTOSAVE", "DEFAULT", "INFERENCE", "PREDICTION", "REVIEW"]


@dataclass
class LabelFilters:
    """Label filters for running a label search."""

    project_id: "ProjectId"
    asset: Optional["AssetFilters"] = None
    author_in: Optional[ListOrTuple[str]] = None
    consensus_mark_gte: Optional[float] = None
    consensus_mark_lte: Optional[float] = None
    created_at: Optional[str] = None
    created_at_gte: Optional[str] = None
    created_at_lte: Optional[str] = None
    honeypot_mark_gte: Optional[float] = None
    honeypot_mark_lte: Optional[float] = None
    id: Optional[LabelId] = None
    id_in: Optional[ListOrTuple[LabelId]] = None
    labeler_in: Optional[ListOrTuple[str]] = None
    reviewer_in: Optional[ListOrTuple[str]] = None
    search: Optional[str] = None
    type_in: Optional[ListOrTuple[LabelType]] = None
    user: Optional["UserFilter"] = None



"""GraphQL Asset operations."""


def get_models_query(fragment: str) -> str:
    """Return the GraphQL projectModels query."""
    return f"""
          query Models($where: ModelWhere!, $first: PageSize!, $skip: Int!) {{
            data: models(where: $where, first: $first, skip: $skip) {{
                {fragment}
              }}
            }}
        """


def get_model_query(fragment: str) -> str:
    """Return the GraphQL model query by ID."""
    return f"""
        query Model($modelId: ID!) {{
            model(id: $modelId) {{
                {fragment}
            }}
        }}
    """


def get_create_model_mutation(fragment: str) -> str:
    """Return the GraphQL createProjectModel mutation."""
    return f"""
          mutation CreateModel($input: CreateModelInput!) {{
            createModel(input: $input) {{
              {fragment}
            }}
          }}
        """


def get_update_model_mutation(fragment: str) -> str:
    """Return the GraphQL updateModel mutation."""
    return f"""
        mutation UpdateModel($id: ID!, $input: UpdateModelInput!) {{
            updateModel(id: $id, input: $input) {{
                {fragment}
            }}
        }}
    """


def get_delete_model_mutation() -> str:
    """Return the GraphQL deleteOrganizationModel mutation."""
    return """
    mutation DeleteModel($deleteModelId: ID!) {
        deleteModel(id: $deleteModelId)
    }
    """


def get_create_project_model_mutation(fragment: str) -> str:
    """Return the GraphQL createProjectModel mutation."""
    return f"""
          mutation CreateProjectModel($input: CreateProjectModelInput!) {{
            createProjectModel(input: $input) {{
              {fragment}
            }}
          }}
        """


def get_update_project_model_mutation(fragment: str) -> str:
    """Return the GraphQL updateProjectModel mutation."""
    return f"""
        mutation UpdateProjectModel($updateProjectModelId: ID!, $input: UpdateProjectModelInput!) {{
            updateProjectModel(id: $updateProjectModelId, input: $input) {{
                {fragment}
            }}
        }}
    """


def get_delete_project_model_mutation() -> str:
    """Return the GraphQL deleteProjectModel mutation."""
    return """
    mutation DeleteProjectModel($deleteProjectModelId: ID!) {
       deleteProjectModel(id: $deleteProjectModelId)
    }
    """


def get_project_models_query(fragment: str) -> str:
    """Return the GraphQL projectModels query."""
    return f"""
        query ProjectModels($where: ProjectModelWhere!, $first: PageSize!, $skip: Int!) {{
          data: projectModels(where: $where, first: $first, skip: $skip) {{
            {fragment}
          }}
        }}
        """


def get_create_llm_asset_mutation(fragment: str) -> str:
    """Return the GraphQL createLLMAsset mutation."""
    return f"""
        mutation CreateLLMAsset($where: ProjectWhere!, $data: CreateLLMAssetData!) {{
            createLLMAsset(where: $where, data: $data) {{
                {fragment}
            }}
        }}
    """


def get_create_chat_item_mutation(fragment: str) -> str:
    """Return the GraphQL createChatItem mutation."""
    return f"""
        mutation CreateChatItem($data: CreateChatItemData!, $where: AssetWhere!) {{
            createChatItem(data: $data, where: $where) {{
                {fragment}
            }}
        }}
    """



"""Base class for entrypoints dealing with GraphQL operations."""

import abc
from typing import List, Optional, Type, TypeVar, cast

from kili.adapters.http_client import HttpClient
from kili.adapters.kili_api_gateway.kili_api_gateway import KiliAPIGateway
from kili.core.graphql.graphql_client import GraphQLClient
from kili.core.helpers import format_result
from kili.domain.asset import AssetExternalId, AssetId
from kili.domain.project import ProjectId
from kili.domain.types import ListOrTuple
from kili.use_cases.asset.utils import AssetUseCasesUtils

T = TypeVar("T")


class BaseOperationEntrypointMixin(abc.ABC):
    """Base class for entrypoints dealing with GraphQL operations."""

    # FIXME: graphql_client and http_client should be removed once
    # all methods have been moved to the new architecture
    graphql_client: GraphQLClient
    http_client: HttpClient
    kili_api_gateway: KiliAPIGateway

    def format_result(self, name: str, result: dict, object_: Optional[Type[T]] = None) -> T:
        """Format the result of a graphQL query.

        FIXME: this should not be used at that level.
        """
        return format_result(name, result, object_, self.http_client)

    def _resolve_asset_ids(
        self,
        asset_ids: Optional[List[str]],
        external_ids: Optional[List[str]],
        project_id: Optional[str],
    ) -> ListOrTuple[AssetId]:
        # FIXME: temporary method to be removed once all methods have been
        # moved to the new architecture
        return AssetUseCasesUtils(self.kili_api_gateway).get_asset_ids_or_throw_error(
            cast(Optional[List[AssetId]], asset_ids),
            cast(Optional[List[AssetExternalId]], external_ids),
            ProjectId(project_id) if project_id else None,
        )



"""GraphQL Tag operations."""

GQL_CHECK_TAG = """
mutation checkTag($data: CheckedTagData!) {
    data: checkTag(data: $data) {
        id
    }
}
"""

GQL_UNCHECK_TAG = """
mutation uncheckTag($projectId: ID!, $tagId: ID!) {
    data: uncheckTag(projectId: $projectId, tagId: $tagId) {
        id
    }
}
"""

GQL_UPDATE_TAG = """
mutation updateTag($tagId: ID!, $data: UpdateTagData!) {
    data: updateTag(tagId: $tagId, data: $data) {
        affectedRows
        updatedTag {
            id
        }
    }
}
"""

GQL_DELETE_TAG = """
mutation deleteTag($tagId: ID!) {
    data: deleteTag(tagId: $tagId)
}
"""


def get_list_tags_by_org_query(fragment: str) -> str:
    """Return the GraphQL query to list tags by organization."""
    return f"""
    query listTagsByOrg {{
        data: listTagsByOrg {{
            {fragment}
        }}
    }}
    """


def get_list_tags_by_project_query(fragment: str) -> str:
    """Return the GraphQL query to list tags by project."""
    return f"""
    query listTagsByProject($projectId: ID!) {{
        data: listTagsByProject(projectId: $projectId) {{
            {fragment}
        }}
    }}
    """



"""GraphQL payload data mappers for api keys operations."""

from typing import Dict

from kili.domain.llm import (
    AzureOpenAICredentials,
    ModelToCreateInput,
    ModelToUpdateInput,
    ModelType,
    OpenAISDKCredentials,
    OrganizationModelFilters,
    ProjectModelFilters,
    ProjectModelToCreateInput,
    ProjectModelToUpdateInput,
)


def model_where_wrapper(filter: OrganizationModelFilters) -> Dict:
    """Build the GraphQL ProjectMapperWhere variable to be sent in an operation."""
    return {
        "organizationId": filter.organization_id,
    }


def project_model_where_mapper(filter: ProjectModelFilters) -> Dict:
    """Build the GraphQL ProjectMapperWhere variable to be sent in an operation."""
    return {
        "projectId": filter.project_id,
        "modelId": filter.model_id,
    }


def map_create_model_input(data: ModelToCreateInput) -> Dict:
    """Build the GraphQL ModelInput variable to be sent in an operation."""
    if data.type == ModelType.AZURE_OPEN_AI and isinstance(
        data.credentials, AzureOpenAICredentials
    ):
        credentials = {
            "apiKey": data.credentials.api_key,
            "deploymentId": data.credentials.deployment_id,
            "endpoint": data.credentials.endpoint,
        }
    elif data.type == ModelType.OPEN_AI_SDK and isinstance(data.credentials, OpenAISDKCredentials):
        credentials = {"apiKey": data.credentials.api_key, "endpoint": data.credentials.endpoint}
    else:
        raise ValueError(
            f"Unsupported model type or credentials: {data.type}, {type(data.credentials)}"
        )

    return {
        "credentials": credentials,
        "name": data.name,
        "type": data.type.value,
        "organizationId": data.organization_id,
    }


def map_update_model_input(data: ModelToUpdateInput) -> Dict:
    """Build the GraphQL UpdateModelInput variable to be sent in an operation."""
    input_dict = {}
    if data.name is not None:
        input_dict["name"] = data.name

    if data.credentials is not None:
        if isinstance(data.credentials, AzureOpenAICredentials):
            credentials = {
                "apiKey": data.credentials.api_key,
                "deploymentId": data.credentials.deployment_id,
                "endpoint": data.credentials.endpoint,
            }
        elif isinstance(data.credentials, OpenAISDKCredentials):
            credentials = {
                "apiKey": data.credentials.api_key,
                "endpoint": data.credentials.endpoint,
            }
        else:
            raise ValueError(f"Unsupported credentials type: {type(data.credentials)}")
        input_dict["credentials"] = credentials

    return input_dict


def map_create_project_model_input(data: ProjectModelToCreateInput) -> Dict:
    """Build the GraphQL ModelInput variable to be sent in an operation."""
    return {
        "projectId": data.project_id,
        "modelId": data.model_id,
        "configuration": data.configuration,
    }


def map_update_project_model_input(data: ProjectModelToUpdateInput) -> Dict:
    """Build the GraphQL UpdateProjectModelInput variable to be sent in an operation."""
    input_dict = {}
    if data.configuration is not None:
        input_dict["configuration"] = data.configuration
    return input_dict


def map_delete_model_input(model_id: str) -> Dict:
    """Map the input for the GraphQL deleteModel mutation."""
    return {
        "deleteModelId": model_id,
    }


def map_delete_project_model_input(project_model_id: str) -> Dict:
    """Map the input for the GraphQL deleteProjectModel mutation."""
    return {
        "deleteProjectModelId": project_model_id,
    }


def map_create_llm_asset_input(data: Dict) -> Dict:
    """Map the input for the createLLMAsset mutation."""
    result = {
        "authorId": data["author_id"],
    }
    if "status" in data:
        result["status"] = data["status"]
    if "label_type" in data:
        result["labelType"] = data["label_type"]
    return result


def map_project_where(project_id: str) -> Dict:
    """Map the 'where' parameter for mutations that require a ProjectWhere."""
    return {"id": project_id}


def map_create_chat_item_input(label_id: str, prompt: str) -> Dict:
    """Map the input for the createChatItem mutation."""
    return {"content": prompt, "role": "USER", "labelId": label_id}


def map_asset_where(asset_id: str) -> Dict:
    """Map the 'where' parameter for the createChatItem mutation."""
    return {"id": asset_id}



"""Service for exporting kili objects."""

from typing import Dict, List, Optional, Union

from kili.adapters.kili_api_gateway.kili_api_gateway import KiliAPIGateway
from kili.domain.asset.asset import AssetFilters
from kili.domain.project import ProjectId

from .dynamic import LLMDynamicExporter
from .static import LLMStaticExporter


def export(  # pylint: disable=too-many-arguments, too-many-locals
    kili_api_gateway: KiliAPIGateway,
    project_id: ProjectId,
    asset_filter: AssetFilters,
    disable_tqdm: Optional[bool],
) -> Optional[List[Dict[str, Union[List[str], str]]]]:
    """Export the selected assets with their labels into the required format, and save it into a file archive."""
    project = kili_api_gateway.get_project(project_id, ["id", "inputType", "jsonInterface"])
    input_type = project["inputType"]

    if input_type == "LLM_RLHF":
        return LLMStaticExporter(kili_api_gateway, disable_tqdm).export(
            project_id, asset_filter, project["jsonInterface"]
        )
    if input_type == "LLM_INSTR_FOLLOWING":
        asset_filter.status_in = ["LABELED", "REVIEWED", "TO_REVIEW"]
        return LLMDynamicExporter(kili_api_gateway, disable_tqdm).export(
            asset_filter, project["jsonInterface"]
        )
    raise ValueError(f'Project Input type "{input_type}" cannot be used for llm exports.')



"""Handle LLM_INSTR_FOLLOWING project exports."""

import logging
from typing import Dict, List, Optional, Union

from kili.adapters.kili_api_gateway.helpers.queries import QueryOptions
from kili.adapters.kili_api_gateway.kili_api_gateway import KiliAPIGateway
from kili.domain.asset.asset import AssetFilters

CHAT_ITEMS_NEEDED_FIELDS = [
    "id",
    "content",
    "createdAt",
    "modelId",
    "parentId",
    "role",
]

LABELS_NEEDED_FIELDS = [
    "annotations.id",
    "author.id",
    "author.email",
    "author.firstname",
    "author.lastname",
    *(f"chatItems.{field}" for field in CHAT_ITEMS_NEEDED_FIELDS),
    "createdAt",
    "id",
    "isLatestLabelForUser",
    "jsonResponse",  # This is needed to keep annotations
    "labelType",
    "modelName",
]

ASSET_NEEDED_FIELDS = [
    "assetProjectModels.id",
    "assetProjectModels.projectModelId",
    "assetProjectModels.configuration",
    "content",
    "externalId",
    "jsonMetadata",
    *(f"labels.{field}" for field in LABELS_NEEDED_FIELDS),
    "status",
]


class LLMDynamicExporter:
    """Handle exports of LLM_RLHF projects."""

    def __init__(self, kili_api_gateway: KiliAPIGateway, disable_tqdm: Optional[bool]):
        self.kili_api_gateway = kili_api_gateway
        self.disable_tqdm = disable_tqdm

    def export(
        self, asset_filter: AssetFilters, json_interface: Dict
    ) -> List[Dict[str, Union[List[str], str]]]:
        """Asset content depends of each label."""
        options = QueryOptions(disable_tqdm=self.disable_tqdm)
        assets = self.kili_api_gateway.list_assets(asset_filter, ASSET_NEEDED_FIELDS, options)
        export_res = []
        for asset in assets:
            # obfuscate models here
            obfuscated_models = {}
            for index, asset_project_model in enumerate(asset["assetProjectModels"]):
                obfuscated_models[asset_project_model["id"]] = f"{chr(65 + index)}"
            for label in asset["labels"]:
                result = {}
                chat_items = label["chatItems"]
                annotations = label["annotations"]
                rounds = self._build_rounds(chat_items, annotations, json_interface)
                for step, round in enumerate(rounds):
                    raw_data = _format_raw_data(
                        round["context"]
                        + round["pre_prompts"]
                        + [round["prompt"]]
                        + round["completion"],
                        label["id"],
                        obfuscated_models,
                    )
                    result[f"{step}"] = {
                        "external_id": asset["externalId"],
                        "metadata": asset["jsonMetadata"],
                        "models": _format_models_object(
                            asset["assetProjectModels"], obfuscated_models
                        ),
                        "labels": [
                            {
                                "author": label["author"]["email"],
                                "created_at": label["createdAt"],
                                "label_type": label["labelType"],
                                "label": _format_json_response(
                                    json_interface["jobs"],
                                    round["annotations"],
                                    round["completion"],
                                    obfuscated_models,
                                ),
                            }
                        ],
                        "raw_data": raw_data,
                        "status": asset["status"],
                    }
                export_res.append(result)
        return export_res

    def _get_round_winner(self, completions, annotations, json_interface):
        """Get the winner completion.

        Find the first comparison job in the json_interface and return the completion that has an annotation of this job targeting it.
        """
        comparison_job = None
        comparison_job_name = None
        for job_name, job in json_interface["jobs"].items():
            if job["mlTask"] == "COMPARISON":
                comparison_job = job
                comparison_job_name = job_name
                break
        if comparison_job is None:
            raise ValueError("No comparison job found in json_interface")

        winnerId = None
        for annotation in annotations:
            if annotation["job"] == comparison_job_name:
                winnerId = annotation["annotationValue"]["choice"]["firstId"]

        if winnerId is None:
            raise ValueError(
                f"Job {comparison_job_name} is used to create context, but no related annotations found"
            )
        res = list(filter(lambda completion: completion["id"] == winnerId, completions))
        return res[0] if len(res) > 0 else None

    def _init_round(self, context):
        return {
            "context": context,
            "prompt": None,
            "pre_prompts": [],
            "completion": [],
            "annotations": [],
        }

    def _build_rounds(self, chat_items, annotations, json_interface):
        """A round is composed of a prompt with n pre-prompts and n completions."""
        ordered_chat_items = sorted(chat_items, key=lambda x: x["createdAt"])
        rounds = []
        current_round = self._init_round([])
        for chat_item in ordered_chat_items:
            role = chat_item["role"].lower() if chat_item["role"] else None
            if role == "user" or role == "system":
                if current_round["prompt"] is not None:
                    rounds.append(current_round)
                    new_context = (
                        current_round["context"]
                        + current_round["pre_prompts"]
                        + [
                            current_round["prompt"],
                            self._get_round_winner(
                                current_round["completion"],
                                current_round["annotations"],
                                json_interface,
                            ),
                        ]
                    )
                    current_round = self._init_round(new_context)

            if role == "user":
                current_round["prompt"] = chat_item
            elif role == "system":
                current_round["pre_prompt"].append(chat_item)
            elif role == "assistant":
                current_round["completion"].append(chat_item)
            else:
                raise ValueError(f"Role {chat_item['role']} not supported")
            current_round["annotations"] += [
                annotation
                for annotation in annotations
                if annotation["chatItemId"] == chat_item["id"]
            ]
        rounds.append(current_round)
        return rounds


def _format_classification_annotation(annotation):
    return annotation["annotationValue"]["categories"]


def _format_transcription_annotation(annotation):
    return annotation["annotationValue"]["text"]


def _format_comparison_annotation(annotation, completions, job, obfuscated_models):
    """Return A_X or B_X depending of the evaluation completion and its score."""
    model_id = None
    for completion in completions:
        if annotation["annotationValue"]["choice"]["firstId"] == completion["id"]:
            model_id = completion["modelId"]
            break

    if model_id is None:
        raise ValueError(f"Failed to found model of annotation {annotation['id']}")

    iteration = 0
    for _comparison_code, comparison_note in job["content"]["options"].items():
        iteration += 1
        if comparison_note["name"] == annotation["annotationValue"]["choice"]["code"]:
            break

    return f"{obfuscated_models[model_id]}_{iteration}"


def _format_json_response(
    jobs_config: Dict, annotations: List[Dict], completions: List[Dict], obfuscated_models: Dict
) -> Dict[str, Union[str, List[str]]]:
    result = {}
    for annotation in annotations:
        formatted_response = None
        job = jobs_config[annotation["job"]]
        if job["mlTask"] == "CLASSIFICATION":
            formatted_response = _format_classification_annotation(annotation)
        elif job["mlTask"] == "TRANSCRIPTION":
            formatted_response = _format_transcription_annotation(annotation)
        elif job["mlTask"] == "COMPARISON":
            formatted_response = _format_comparison_annotation(
                annotation, completions, job, obfuscated_models
            )

        if formatted_response is None:
            logging.warning(
                f"Annotation with job {annotation['job']} with mlTask {job['mlTask']} not supported. Ignored in the export."
            )
        else:
            result[annotation["job"]] = formatted_response

    return result


def _format_raw_data(chat_items, label_id, obfuscated_models):
    raw_data = []
    for chat_item in chat_items:
        raw_data.append(
            {
                "content": chat_item["content"],
                "role": chat_item["role"].lower(),
                "chat_id": label_id,
                "id": chat_item["id"],
                "model": obfuscated_models[chat_item["modelId"]]
                if chat_item["modelId"] in obfuscated_models
                else None,
            }
        )
    return raw_data


def _format_models_object(models, obfuscated_models):
    res = {}
    for model in models:
        model_configuration = {
            "id": model["id"],
            "projectModelId": model["projectModelId"],
            "configuration": model["configuration"],
        }
        res[obfuscated_models[model["id"]]] = model_configuration

    return res



"""Geojson linestring utilities."""

from typing import Any, Dict, List, Optional


def kili_line_to_geojson_linestring(polyline: List[Dict[str, float]]) -> Dict[str, Any]:
    """Convert a Kili line to a geojson linestring.

    Args:
        polyline: a Kili line (polyline).

    Returns:
        A geojson linestring.

    !!! Example
        ```python
        >>> polyline = [{"x": 1.0, "y": 2.0}, {"x": 3.0, "y": 4.0}]
        >>> kili_line_to_geojson_linestring(polyline)
        {
            "type": "LineString",
            "coordinates": [[1.0, 2.0], [3.0, 4.0]]
        }
        ```
    """
    ret = {"type": "LineString", "coordinates": []}
    ret["coordinates"] = [[vertex["x"], vertex["y"]] for vertex in polyline]
    return ret  # type: ignore


def kili_line_annotation_to_geojson_linestring_feature(
    polyline_annotation: Dict[str, Any], job_name: Optional[str] = None
) -> Dict[str, Any]:
    """Convert a Kili line annotation to a geojson linestring feature.

    Args:
        polyline_annotation: a Kili line annotation.
        job_name: the name of the job to which the annotation belongs.

    Returns:
        A geojson linestring feature.

    !!! Example
        ```python
        >>> polyline = {
            'children': {},
            'polyline': [{'x': -79.0, 'y': -3.0}, {'x': -79.0, 'y': -3.0}],
            'categories': [{'name': 'A'}],
            'mid': 'mid_object',
            'type': 'polyline'
        }
        >>> kili_line_annotation_to_geojson_linestring_feature(polyline, 'job_name')
        {
            'type': 'Feature',
            'geometry': {
                'type': 'LineString',
                'coordinates': [[-79.0, -3.0], [-79.0, -3.0]]},
                'id': 'mid_object',
                'properties': {
                    'kili': {
                        'categories': [{'name': 'A'}],
                        'children': {},
                        'job': 'job_name'
                    }
                }
        }
        ```
    """
    assert (
        polyline_annotation["type"] == "polyline"
    ), f"Annotation type must be `polyline`, got: {polyline_annotation['type']}"

    ret = {
        "type": "Feature",
        "geometry": kili_line_to_geojson_linestring(polyline_annotation["polyline"]),
    }
    if "mid" in polyline_annotation:
        ret["id"] = polyline_annotation["mid"]
    ret["properties"] = {
        "kili": {k: v for k, v in polyline_annotation.items() if k not in ["mid", "polyline"]}
    }
    if job_name is not None:
        ret["properties"]["kili"]["job"] = job_name
    return ret


def geojson_linestring_feature_to_kili_line_annotation(
    line: Dict[str, Any],
    categories: Optional[List[Dict]] = None,
    children: Optional[Dict] = None,
    mid: Optional[str] = None,
) -> Dict[str, Any]:
    # pylint: disable=line-too-long
    """Convert a geojson linestring feature to a Kili line annotation.

    Args:
        line: a geojson linestring feature.
        categories: the categories of the annotation.
            If not provided, the categories are taken from the `kili` key of the geojson feature properties.
        children: the children of the annotation.
            If not provided, the children are taken from the `kili` key of the geojson feature properties.
        mid: the mid of the annotation.
            If not provided, the mid is taken from the `id` key of the geojson feature.

    Returns:
        A Kili line annotation.

    !!! Example
        ```python
        >>> line = {
            'type': 'Feature',
            'geometry': {
                'type': 'LineString',
                'coordinates': [[-79.0, -3.0], [-79.0, -3.0]]},
            }
            'id': 'mid_object',
            'properties': {
                'kili': {
                    'categories': [{'name': 'A'}],
                    'children': {},
                    'job': 'job_name'
                }
            }
        }
        >>> geojson_linestring_feature_to_kili_line_annotation(line)
        {
            'children': {},
            'polyline': [{'x': -79.0, 'y': -3.0}, {'x': -79.0, 'y': -3.0}],
            'categories': [{'name': 'A'}],
            'mid': 'mid_object',
            'type': 'polyline'
        }
        ```
    """
    assert line["type"] == "Feature", f"Feature type must be `Feature`, got: {line['type']}"
    assert (
        line["geometry"]["type"] == "LineString"
    ), f"Geometry type must be `LineString`, got: {line['geometry']['type']}"

    children = children or line["properties"].get("kili", {}).get("children", {})
    categories = categories or line["properties"]["kili"]["categories"]

    ret = {
        "children": children,
        "categories": categories,
        "type": "polyline",
    }
    ret["polyline"] = [{"x": coord[0], "y": coord[1]} for coord in line["geometry"]["coordinates"]]

    if mid is not None:
        ret["mid"] = str(mid)
    elif "id" in line:
        ret["mid"] = str(line["id"])

    return ret



"""Helpers to create point annotations."""

from typing import Dict, Literal, Optional, Union


def point_to_normalized_point(
    point: Dict[str, Union[int, float]],
    img_width: Optional[Union[int, float]] = None,
    img_height: Optional[Union[int, float]] = None,
    origin_location: Literal["top_left", "bottom_left"] = "bottom_left",
) -> Dict[Literal["x", "y"], float]:
    # pylint: disable=line-too-long
    """Converts a 2D point to a Kili normalized vertex.

    The output can be used to create object detection annotations. See the [documentation](https://docs.kili-technology.com/reference/export-object-entity-detection-and-relation) for more details.

    A point is a dict with keys `"x"` and `"y"`, and corresponding values in pixels (`int` or `float`).

    Conventions for the input point:

    - The origin is defined by the `origin_location` argument.
    - x-axis is horizontal and goes from left to right.
    - y-axis is vertical. If `origin_location` is `"top_left"`, it goes from top to bottom. If `origin_location` is `"bottom_left"`, it goes from bottom to top.

    Conventions for the output point:

    - The origin is the top left corner of the image.
    - x-axis is horizontal and goes from left to right.
    - y-axis is vertical and goes from top to bottom.

    If the image width and height are provided, the input point coordinates will be normalized to `[0, 1]`.
    If not, the method expects the input point coordinates to be already normalized.

    Args:
        point: Point to convert.
        img_width: Width of the image the point is defined in.
        img_height: Height of the image the point is defined in.
        origin_location: Location of the origin of input point coordinate system. Can be either `top_left` or `bottom_left`.

    Returns:
        A dict with keys `"x"` and `"y"`, and corresponding normalized values.

    !!! Example
        ```python
        from kili.utils.labels.point import point_to_normalized_point

        normalized_point = point_to_normalized_point({"x": 5, "y": 40}, img_width=100, img_height=100)

        json_response = {
            "OBJECT_DETECTION_JOB": {
                "annotations": [
                    {
                        "point": normalized_point,
                        "categories": [{"name": "CLASS_A"}],
                        "type": "marker",
                    }
                ]
            }
        }
        ```
    """
    if (img_width is None) != (img_height is None):
        raise ValueError("img_width and img_height must be both None or both not None.")

    if img_width is not None and img_height is not None:
        point = {
            "x": point["x"] / img_width,
            "y": point["y"] / img_height,
        }

    if origin_location == "bottom_left":
        point = {"x": point["x"], "y": 1 - point["y"]}

    assert 0 <= point["x"] <= 1, f"Point x coordinate {point['x']} should be in [0, 1]."
    assert 0 <= point["y"] <= 1, f"Point y coordinate {point['y']} should be in [0, 1]."

    return {"x": point["x"], "y": point["y"]}


def normalized_point_to_point(
    point: Dict[str, float],
    img_width: Optional[Union[int, float]] = None,
    img_height: Optional[Union[int, float]] = None,
    origin_location: Literal["top_left", "bottom_left"] = "bottom_left",
) -> Dict[Literal["x", "y"], float]:
    # pylint: disable=line-too-long
    """Convert a Kili normalized vertex to a 2D point.

    It is the inverse of the method `point_to_normalized_point`.

    A point is a dict with keys `"x"` and `"y"`, and corresponding values in pixels (`int` or `float`).

    Conventions for the input point:

    - The origin is the top left corner of the image.
    - x-axis is horizontal and goes from left to right.
    - y-axis is vertical and goes from top to bottom.

    Conventions for the output point:

    - The origin is defined by the `origin_location` argument.
    - x-axis is horizontal and goes from left to right.
    - y-axis is vertical. If `origin_location` is `"top_left"`, it goes from top to bottom. If `origin_location` is `"bottom_left"`, it goes from bottom to top.

    If the image width and height are provided, the output point coordinates will be scaled to the image size.
    If not, the method will return a point with normalized coordinates.

    Args:
        point: Point to convert.
        img_width: Width of the image the point is defined in.
        img_height: Height of the image the point is defined in.
        origin_location: Location of the origin of output point coordinate system. Can be either `top_left` or `bottom_left`.

    Returns:
        A dict with keys `"x"` and `"y"`, and corresponding values in pixels.
    """
    if (img_width is None) != (img_height is None):
        raise ValueError("img_width and img_height must be both None or both not None.")

    if origin_location == "bottom_left":
        point = {"x": point["x"], "y": 1 - point["y"]}

    img_height = img_height or 1
    img_width = img_width or 1

    return {"x": point["x"] * img_width, "y": point["y"] * img_height}



"""Base class for all KiliAPIGateway Operation Mixin classes."""

from abc import ABC

from kili.adapters.http_client import HttpClient
from kili.core.graphql.graphql_client import GraphQLClient


class BaseOperationMixin(ABC):
    """Base class for all KiliAPIGateway Operation Mixin classes.

    It is used to share the GraphQL client between all methods classes.

    It is not meant to be used and instantiated directly.
    """

    graphql_client: GraphQLClient  # instantiated in the KiliAPIGateway child class
    http_client: HttpClient  # instantiated in the KiliAPIGateway child class



"""Asset mutations."""

import warnings
from typing import Any, Dict, List, Literal, Optional, Union, cast

from tenacity import retry
from tenacity.retry import retry_if_exception_type
from tenacity.wait import wait_exponential
from typeguard import typechecked

from kili.adapters.kili_api_gateway.helpers.queries import QueryOptions
from kili.core.helpers import is_empty_list_with_warning
from kili.core.utils.pagination import mutate_from_paginated_call
from kili.domain.asset import AssetFilters
from kili.domain.project import ProjectId
from kili.entrypoints.base import BaseOperationEntrypointMixin
from kili.entrypoints.mutations.asset.helpers import (
    process_update_properties_in_assets_parameters,
)
from kili.entrypoints.mutations.asset.queries import (
    GQL_ADD_ALL_LABELED_ASSETS_TO_REVIEW,
    GQL_DELETE_MANY_FROM_DATASET,
    GQL_SEND_BACK_ASSETS_TO_QUEUE,
    GQL_UPDATE_PROPERTIES_IN_ASSETS,
)
from kili.entrypoints.mutations.exceptions import MutationError
from kili.exceptions import MissingArgumentError
from kili.services.asset_import import import_assets
from kili.services.asset_import_csv import get_text_assets_from_csv
from kili.utils.assets import PageResolution
from kili.utils.logcontext import for_all_methods, log_call


@for_all_methods(log_call, exclude=["__init__"])
class MutationsAsset(BaseOperationEntrypointMixin):
    """Set of Asset mutations."""

    # pylint: disable=too-many-arguments,too-many-locals
    @typechecked
    def append_many_to_dataset(
        self,
        project_id: str,
        content_array: Optional[Union[List[str], List[dict], List[List[dict]]]] = None,
        multi_layer_content_array: Optional[List[List[dict]]] = None,
        external_id_array: Optional[List[str]] = None,
        id_array: Optional[List[str]] = None,
        is_honeypot_array: Optional[List[bool]] = None,
        status_array: Optional[List[str]] = None,
        json_content_array: Optional[List[Union[List[Union[dict, str]], None]]] = None,
        json_metadata_array: Optional[List[dict]] = None,
        disable_tqdm: Optional[bool] = None,
        wait_until_availability: bool = True,
        from_csv: Optional[str] = None,
        csv_separator: str = ",",
    ) -> Dict[Literal["id", "asset_ids"], Union[str, List[str]]]:
        # pylint: disable=line-too-long
        """Append assets to a project.

        Args:
            project_id: Identifier of the project
            content_array: List of elements added to the assets of the project
                Must not be None except if you provide multi_layer_content_array or json_content_array.

                - For a `TEXT` project, the content can be either raw text, or URLs to TEXT assets.
                - For an `IMAGE` / `PDF` project, the content can be either URLs or paths to existing
                    images/pdf on your computer.
                - For a VIDEO project, the content can be either URLs pointing to videos hosted on a web server or paths to
                existing video files on your computer. If you want to import video from frames, look at the json_content
                section below.
                - For an `VIDEO_LEGACY` project, the content can be only be URLs.
                - For an `LLM_RLHF` project, the content can be dicts with the keys `prompt` and `completions`,
                paths to local json files or URLs to json files.
            multi_layer_content_array: List containing multiple lists of paths.
                Each path correspond to a layer of a geosat asset. Should be used only for `IMAGE` projects.
            external_id_array: List of external ids given to identify the assets.
                If None, random identifiers are created.
            id_array: Disabled parameter. Do not use.
            is_honeypot_array:  Whether to use the asset for honeypot
            status_array: DEPRECATED and does not have any effect.
            json_content_array: Useful for `VIDEO` or `TEXT` or `IMAGE` projects only.

                - For `VIDEO` projects, each element is a sequence of frames, i.e. a
                    list of URLs to images or a list of paths to images.
                - For `TEXT` projects, each element is a json_content dict,
                    formatted according to documentation [on how to import
                rich-text assets](https://python-sdk-docs.kili-technology.com/latest/sdk/tutorials/import_text_assets/).
                - For `IMAGES` projects, it is used for satellite imagery each element is a list of json_content dicts
                    formatted according to documentation [on how to add multi-layer images]
                    (https://docs.kili-technology.com/docs/adding-assets-to-project#adding-multi-layer-images)

            json_metadata_array: The metadata given to each asset should be stored in a json like dict with keys.

                - Add metadata visible on the asset with the following keys: `imageUrl`, `text`, `url`.
                    Example for one asset: `json_metadata_array = [{'imageUrl': '','text': '','url': ''}]`.
                - For VIDEO projects (and not VIDEO_LEGACY), you can specify a value with key 'processingParameters' to specify the sampling rate (default: 30).
                    Example for one asset: `json_metadata_array = [{'processingParameters': {'framesPlayedPerSecond': 10}}]`.
                - In Image projects with geoTIFF assets, you can specify the epsg, the `minZoom` and `maxZoom` values for the `processingParameters` key.
                    - The epsg is a number that defines the projection that will be used for the asset. Values that can be used are either 4326 or 3857, the 2
                    projections that we support. If this number is not set, by default we keep the initial projection of the asset if it is 4326 or 3857, either
                    we reproject the asset to EPSG:3857 by default.
                    - The `minZoom` parameter defines the zoom level that users are not allowed to zoom out from. It also affects the zoom levels for which we
                    generate the tiles when tiling the asset (for asset with size > 30MB).
                    - The `maxZoom` value affects asset generation: the higher the value, the greater the level of details and the size of the asset. It also affects
                    the zoom levels for which we generate the tiles when tiling the asset (for asset with size > 30MB).
                    - Example for one asset: `json_metadata_array = [{'processingParameters': {'epsg': 3758, 'minZoom': 17, 'maxZoom': 19}}]`.
            disable_tqdm: If `True`, the progress bar will be disabled
            wait_until_availability: If `True`, the function will return once the assets are fully imported in Kili.
                If `False`, the function will return faster but the assets might not be fully processed by the server.
            from_csv: Path to a csv file containing the text assets to import.
                Only used for `TEXT` projects.
                If provided, `content_array` and `external_id_array` must be None.
                The csv file header must specify the columns `content` and `externalId`.
            csv_separator: Separator used in the csv file. Only used if `from_csv` is provided.


        Returns:
            A dictionary with two fields: `id` which is the project id and `asset_ids` which is a list of the created asset ids.
            In the case where assets are uploaded asynchronously (for video imported as frames or big images or tiff images), the method return an empty list of asset ids.

        Examples:
            >>> kili.append_many_to_dataset(
                    project_id=project_id,
                    content_array=['https://upload.wikimedia.org/wikipedia/en/7/7d/Lenna_%28test_image%29.png'])

        !!! example "Recipe"
            - For more detailed examples on how to import assets,
                see [the recipe](https://docs.kili-technology.com/recipes/importing-data).
            - For more detailed examples on how to import text assets,
                see [the recipe](https://python-sdk-docs.kili-technology.com/latest/sdk/tutorials/import_text_assets/).
        """
        if from_csv is not None:
            if content_array is not None or external_id_array is not None:
                raise ValueError(
                    "If from_csv is provided, content_array and external_id_array must not be"
                    " provided."
                )
            content_array, external_id_array = get_text_assets_from_csv(
                from_csv=from_csv, csv_separator=csv_separator
            )

        if (
            is_empty_list_with_warning("append_many_to_dataset", "content_array", content_array)
            or is_empty_list_with_warning(
                "append_many_to_dataset", "json_content_array", json_content_array
            )
            or is_empty_list_with_warning(
                "append_many_to_dataset", "multi_layer_content_array", multi_layer_content_array
            )
        ):
            return {"id": project_id, "asset_ids": []}

        if status_array is not None:
            warnings.warn(
                "status_array is deprecated and will not be sent in the call. Asset status is"
                " automatically computed based on its labels and cannot be overwritten.",
                DeprecationWarning,
                stacklevel=1,
            )

        if (
            content_array is None
            and multi_layer_content_array is None
            and json_content_array is None
        ):
            raise ValueError(
                "Variables content_array, multi_layer_content_array and json_content_array cannot be both None."
            )

        if content_array is not None and multi_layer_content_array is not None:
            raise ValueError(
                "Variables content_array and multi_layer_content_array cannot be both provided."
            )

        nb_data = (
            len(content_array)
            if content_array is not None
            else (
                len(multi_layer_content_array)
                if multi_layer_content_array is not None
                else len(json_content_array)  # type:ignore
            )
        )

        field_mapping = {
            "content": content_array,
            "multi_layer_content": multi_layer_content_array,
            "json_content": json_content_array,
            "external_id": external_id_array,
            "id": id_array,
            "json_metadata": json_metadata_array,
            "is_honeypot": is_honeypot_array,
        }
        assets = [{}] * nb_data
        for key, value in field_mapping.items():
            if value is not None:
                assets = [{**assets[i], key: value[i]} for i in range(nb_data)]
        created_asset_ids = import_assets(
            self,  # pyright: ignore[reportGeneralTypeIssues]
            project_id=ProjectId(project_id),
            assets=assets,
            disable_tqdm=disable_tqdm,
            verify=wait_until_availability,
        )
        return {"id": project_id, "asset_ids": created_asset_ids}

    @typechecked
    def update_properties_in_assets(
        self,
        asset_ids: Optional[List[str]] = None,
        external_ids: Optional[List[str]] = None,
        priorities: Optional[List[int]] = None,
        json_metadatas: Optional[List[Union[dict, str]]] = None,
        consensus_marks: Optional[List[float]] = None,
        honeypot_marks: Optional[List[float]] = None,
        to_be_labeled_by_array: Optional[List[List[str]]] = None,
        contents: Optional[List[str]] = None,
        json_contents: Optional[List[str]] = None,
        status_array: Optional[List[str]] = None,
        is_used_for_consensus_array: Optional[List[bool]] = None,
        is_honeypot_array: Optional[List[bool]] = None,
        project_id: Optional[str] = None,
        resolution_array: Optional[List[Dict]] = None,
        page_resolutions_array: Optional[
            Union[List[List[dict]], List[List[PageResolution]]]
        ] = None,
    ) -> List[Dict[Literal["id"], str]]:
        """Update the properties of one or more assets.

        Args:
            asset_ids: The internal asset IDs to modify.
            external_ids: The external asset IDs to modify (if `asset_ids` is not already provided).
            priorities: You can change the priority of the assets.
                By default, all assets have a priority of 0.
            json_metadatas: The metadata given to an asset should be stored
                in a json like dict with keys `imageUrl`, `text`, `url`:
                `json_metadata = {'imageUrl': '','text': '','url': ''}`
            consensus_marks: Should be between 0 and 1.
            honeypot_marks: Should be between 0 and 1.
            to_be_labeled_by_array: If given, each element of the list should contain the emails of
                the labelers authorized to label the asset.
            contents: - For a NLP project, the content can be directly in text format.
                - For an Image / Video / Pdf project, the content must be hosted on a web server,
                and you point Kili to your data by giving the URLs.
            json_contents: - For a NLP project, the `json_content`
                is a text formatted using RichText.
                - For a Video project, the`json_content` is a json containg urls pointing
                    to each frame of the video.
            status_array: DEPRECATED and does not have any effect.
            is_used_for_consensus_array: Whether to use the asset to compute consensus kpis or not.
            is_honeypot_array: Whether to use the asset for honeypot.
            project_id: The project ID. Only required if `external_ids` argument is provided.
            resolution_array: The resolution of each asset (for image and video assets).
                Each resolution must be passed as a dictionary with keys `width` and `height`.
            page_resolutions_array: The resolution of each page of the asset (for PDF assets).
                Note that each element of the array should contain all the pages resolutions of the
                corresponding asset. Each resolution can be passed as a
                `kili.utils.assets.PageResolution` object, or as a dictionary with keys `width`,
                `height`, `pageNumber` and optionally `rotation`.

        Returns:
            A list of dictionaries with the asset ids.

        Examples:
            >>> kili.update_properties_in_assets(
                    asset_ids=["ckg22d81r0jrg0885unmuswj8", "ckg22d81s0jrh0885pdxfd03n"],
                    consensus_marks=[1, 0.7],
                    contents=[None, 'https://to/second/asset.png'],
                    honeypot_marks=[0.8, 0.5],
                    is_honeypot_array=[True, True],
                    is_used_for_consensus_array=[True, False],
                    priorities=[None, 2],
                    to_be_labeled_by_array=[['test+pierre@kili-technology.com'], None],
                )

                # The following call updates the pages resolutions of PDF assets.
            >>> kili.update_properties_in_assets(
                    asset_ids=["ckg22d81r0jrg0885unmuswj8", "ckg22d81s0jrh0885pdxfd03n"],
                    page_resolutions_array=[
                        [
                            PageResolution(width=480, height=640, page_number=1),
                            PageResolution(width=480, height=640, page_number=2),
                        ],[
                            PageResolution(width=340, height=512, page_number=1),
                            PageResolution(width=680, height=1024, page_number=2, rotation=90),
                            PageResolution(width=680, height=1024, page_number=3),
                        ]
                    ],
                )
        """
        if is_empty_list_with_warning(
            "update_properties_in_assets", "asset_ids", asset_ids
        ) or is_empty_list_with_warning(
            "update_properties_in_assets", "external_ids", external_ids
        ):
            return []

        if status_array is not None:
            warnings.warn(
                "status_array is deprecated and will not be sent in the call. Asset status is"
                " automatically computed based on its labels and cannot be overwritten.",
                DeprecationWarning,
                stacklevel=1,
            )
        if asset_ids is not None and external_ids is not None:
            warnings.warn(
                "The use of `external_ids` argument has changed. It is now used to identify"
                " which properties of which assets to update. Please use"
                " `kili.change_asset_external_ids()` method instead to change asset external"
                " IDs.",
                DeprecationWarning,
                stacklevel=1,
            )
            raise MissingArgumentError("Please provide either `asset_ids` or `external_ids`.")

        resolved_asset_ids = self._resolve_asset_ids(asset_ids, external_ids, project_id)

        properties_to_batch = process_update_properties_in_assets_parameters(
            cast(List[str], resolved_asset_ids),
            priorities=priorities,
            json_metadatas=json_metadatas,
            consensus_marks=consensus_marks,
            honeypot_marks=honeypot_marks,
            to_be_labeled_by_array=to_be_labeled_by_array,
            contents=contents,
            json_contents=json_contents,
            is_used_for_consensus_array=is_used_for_consensus_array,
            is_honeypot_array=is_honeypot_array,
            resolution_array=resolution_array,
            page_resolutions_array=page_resolutions_array,
        )

        def generate_variables(batch: Dict) -> Dict:
            asset_ids = batch.pop("assetId")
            data_array = [dict(zip(batch, t)) for t in zip(*batch.values())]  # type: ignore
            return {
                "whereArray": [{"id": asset_id} for asset_id in asset_ids],
                "dataArray": data_array,
            }

        results = mutate_from_paginated_call(
            self,
            properties_to_batch,
            generate_variables,
            GQL_UPDATE_PROPERTIES_IN_ASSETS,
        )
        formated_results = [self.format_result("data", result, None) for result in results]
        return [item for batch_list in formated_results for item in batch_list]

    @typechecked
    def change_asset_external_ids(
        self,
        new_external_ids: List[str],
        asset_ids: Optional[List[str]] = None,
        external_ids: Optional[List[str]] = None,
        project_id: Optional[str] = None,
    ) -> List[Dict[Literal["id"], str]]:
        """Update the external IDs of one or more assets.

        Args:
            new_external_ids: The new external IDs of the assets.
            asset_ids: The asset IDs to modify.
            external_ids: The external asset IDs to modify (if `asset_ids` is not already provided).
            project_id: The project ID. Only required if `external_ids` argument is provided.

        Returns:
            A list of dictionaries with the asset ids.

        Examples:
            >>> kili.change_asset_external_ids(
                    new_external_ids=["asset1", "asset2"],
                    asset_ids=["ckg22d81r0jrg0885unmuswj8", "ckg22d81s0jrh0885pdxfd03n"],
                )
        """
        if is_empty_list_with_warning(
            "change_asset_external_ids", "new_external_ids", new_external_ids
        ):
            return []

        resolved_asset_ids = self._resolve_asset_ids(asset_ids, external_ids, project_id)

        properties_to_batch = process_update_properties_in_assets_parameters(
            asset_ids=cast(List[str], resolved_asset_ids),
            external_ids=new_external_ids,
        )

        def generate_variables(batch: Dict) -> Dict:
            asset_ids = batch.pop("assetId")
            data_array = [dict(zip(batch, t)) for t in zip(*batch.values())]  # type: ignore
            return {
                "whereArray": [{"id": asset_id} for asset_id in asset_ids],
                "dataArray": data_array,
            }

        results = mutate_from_paginated_call(
            self,
            properties_to_batch,
            generate_variables,
            GQL_UPDATE_PROPERTIES_IN_ASSETS,
        )
        formated_results = [self.format_result("data", result, None) for result in results]
        return [item for batch_list in formated_results for item in batch_list]

    @typechecked
    def delete_many_from_dataset(
        self,
        asset_ids: Optional[List[str]] = None,
        external_ids: Optional[List[str]] = None,
        project_id: Optional[str] = None,
    ) -> Optional[Dict[Literal["id"], str]]:
        """Delete assets from a project.

        Args:
            asset_ids: The list of asset internal IDs to delete.
            external_ids: The list of asset external IDs to delete.
            project_id: The project ID. Only required if `external_ids` argument is provided.

        Returns:
            A dict object with the project `id`.
        """
        if is_empty_list_with_warning(
            "delete_many_from_dataset", "asset_ids", asset_ids
        ) or is_empty_list_with_warning("delete_many_from_dataset", "external_ids", external_ids):
            return None

        resolved_asset_ids = self._resolve_asset_ids(asset_ids, external_ids, project_id)

        properties_to_batch = {"asset_ids": resolved_asset_ids}

        def generate_variables(batch):
            return {"where": {"idIn": batch["asset_ids"]}}

        @retry(
            wait=wait_exponential(multiplier=1, min=1, max=8),
            retry=retry_if_exception_type(MutationError),
            reraise=True,
        )
        def verify_last_batch(last_batch: Dict, results: List) -> None:
            """Check that all assets in the last batch have been deleted."""
            if project_id is not None:
                project_id_ = project_id
            # in some case the results is [{'data': None}]
            elif isinstance(results[0]["data"], Dict) and results[0]["data"].get("id"):
                project_id_ = results[0]["data"].get("id")
            else:
                return

            asset_ids = last_batch["asset_ids"][-1:]  # check last asset of the batch only

            nb_assets_in_kili = self.kili_api_gateway.count_assets(
                AssetFilters(
                    project_id=ProjectId(project_id_),
                    asset_id_in=asset_ids,
                )
            )
            if nb_assets_in_kili > 0:
                raise MutationError("Failed to delete some assets.")

        results = mutate_from_paginated_call(
            self,
            properties_to_batch,
            generate_variables,
            GQL_DELETE_MANY_FROM_DATASET,
            last_batch_callback=verify_last_batch,
        )
        return self.format_result("data", results[0])

    @typechecked
    def add_to_review(
        self,
        asset_ids: Optional[List[str]] = None,
        external_ids: Optional[List[str]] = None,
        project_id: Optional[str] = None,
    ) -> Optional[Dict[str, Any]]:
        """Add assets to review.

        !!! warning
            Assets without any label will be ignored.

        Args:
            asset_ids: The asset internal IDs to add to review.
            external_ids: The asset external IDs to add to review.
            project_id: The project ID. Only required if `external_ids` argument is provided.

        Returns:
            A dict object with the project `id` and the `asset_ids` of assets moved to review.
            `None` if no assets have changed status (already had `TO_REVIEW` status for example).
            An error message if mutation failed.

        Examples:
            >>> kili.add_to_review(
                    asset_ids=[
                        "ckg22d81r0jrg0885unmuswj8",
                        "ckg22d81s0jrh0885pdxfd03n",
                    ],
                )
        """
        if is_empty_list_with_warning(
            "add_to_review", "asset_ids", asset_ids
        ) or is_empty_list_with_warning("add_to_review", "external_ids", external_ids):
            return None

        resolved_asset_ids = self._resolve_asset_ids(asset_ids, external_ids, project_id)

        properties_to_batch = {"asset_ids": resolved_asset_ids}

        def generate_variables(batch):
            return {"where": {"idIn": batch["asset_ids"]}}

        @retry(
            wait=wait_exponential(multiplier=1, min=1, max=8),
            retry=retry_if_exception_type(MutationError),
            reraise=True,
        )
        def verify_last_batch(last_batch: Dict, results: List) -> None:
            """Check that all assets in the last batch have been sent to review."""
            if project_id is not None:
                project_id_ = project_id
            # in some case the results is [{'data': None}]
            elif isinstance(results[0]["data"], Dict) and results[0]["data"].get("id"):
                project_id_ = results[0]["data"].get("id")
            else:
                return

            asset_ids = last_batch["asset_ids"][-1:]  # check last asset of the batch only
            nb_assets_in_review = self.kili_api_gateway.count_assets(
                AssetFilters(
                    project_id=ProjectId(project_id_),
                    asset_id_in=asset_ids,
                    status_in=["TO_REVIEW"],
                )
            )
            if len(asset_ids) != nb_assets_in_review:
                raise MutationError("Failed to send some assets to review")

        results = mutate_from_paginated_call(
            self,
            properties_to_batch,
            generate_variables,
            GQL_ADD_ALL_LABELED_ASSETS_TO_REVIEW,
            last_batch_callback=verify_last_batch,
        )
        result = self.format_result("data", results[0])
        # unlike send_back_to_queue, the add_to_review mutation doesn't always return the project ID
        # it happens when no assets have been sent to review
        if isinstance(result, dict) and "id" in result:
            assets_in_review = self.kili_api_gateway.list_assets(
                AssetFilters(
                    project_id=result["id"],
                    asset_id_in=resolved_asset_ids,
                    status_in=["TO_REVIEW"],
                ),
                ["id"],
                QueryOptions(disable_tqdm=True),
            )
            result["asset_ids"] = [asset["id"] for asset in assets_in_review]
        return result

    @typechecked
    def send_back_to_queue(
        self,
        asset_ids: Optional[List[str]] = None,
        external_ids: Optional[List[str]] = None,
        project_id: Optional[str] = None,
    ) -> Optional[Dict[str, Any]]:
        """Send assets back to queue.

        Args:
            asset_ids: List of internal IDs of assets to send back to queue.
            external_ids: List of external IDs of assets to send back to queue.
            project_id: The project ID. Only required if `external_ids` argument is provided.

        Returns:
            A dict object with the project `id` and the `asset_ids` of assets moved to queue.
            An error message if mutation failed.

        Examples:
            >>> kili.send_back_to_queue(
                    asset_ids=[
                        "ckg22d81r0jrg0885unmuswj8",
                        "ckg22d81s0jrh0885pdxfd03n",
                        ],
                )
        """
        if is_empty_list_with_warning(
            "send_back_to_queue", "asset_ids", asset_ids
        ) or is_empty_list_with_warning("send_back_to_queue", "external_ids", external_ids):
            return None

        resolved_asset_ids = self._resolve_asset_ids(asset_ids, external_ids, project_id)

        properties_to_batch = {"asset_ids": resolved_asset_ids}

        def generate_variables(batch):
            return {"where": {"idIn": batch["asset_ids"]}}

        @retry(
            wait=wait_exponential(multiplier=1, min=1, max=8),
            retry=retry_if_exception_type(MutationError),
            reraise=True,
        )
        def verify_last_batch(last_batch: Dict, results: List) -> None:
            """Check that all assets in the last batch have been sent back to queue."""
            if project_id is not None:
                project_id_ = project_id
            # in some case the results is [{'data': None}]
            elif isinstance(results[0]["data"], Dict) and results[0]["data"].get("id"):
                project_id_ = results[0]["data"].get("id")
            else:
                return

            asset_ids = last_batch["asset_ids"][-1:]  # check lastest asset of the batch only
            nb_assets_in_queue = self.kili_api_gateway.count_assets(
                AssetFilters(
                    project_id=ProjectId(project_id_),
                    asset_id_in=asset_ids,
                    status_in=["ONGOING"],
                )
            )
            if len(asset_ids) != nb_assets_in_queue:
                raise MutationError("Failed to send some assets back to queue")

        results = mutate_from_paginated_call(
            self,
            properties_to_batch,
            generate_variables,
            GQL_SEND_BACK_ASSETS_TO_QUEUE,
            last_batch_callback=verify_last_batch,
        )
        result = self.format_result("data", results[0])
        if isinstance(result, dict) and "id" in result:
            assets_in_queue = self.kili_api_gateway.list_assets(
                AssetFilters(
                    project_id=result["id"],
                    asset_id_in=resolved_asset_ids,
                    status_in=["ONGOING"],
                ),
                ["id"],
                QueryOptions(disable_tqdm=True),
            )
            result["asset_ids"] = [asset["id"] for asset in assets_in_queue]
        return result



"""Handle LLM_RLHF project exports."""

import json
import logging
from pathlib import Path
from typing import Dict, List, Optional, Union

from kili.adapters.kili_api_gateway.helpers.queries import QueryOptions
from kili.adapters.kili_api_gateway.kili_api_gateway import KiliAPIGateway
from kili.domain.asset.asset import AssetFilters
from kili.domain.project import ProjectId
from kili.services.asset_import.helpers import SEPARATOR
from kili.services.export.format.llm.types import ExportLLMItem
from kili.use_cases.asset.media_downloader import MediaDownloader
from kili.utils.tempfile import TemporaryDirectory

ASSET_NEEDED_FIELDS = [
    "content",
    "externalId",
    "jsonMetadata",
    "labels.jsonResponse",
    "labels.author.id",
    "labels.author.email",
    "labels.author.firstname",
    "labels.author.lastname",
    "labels.createdAt",
    "labels.isLatestLabelForUser",
    "labels.labelType",
    "labels.modelName",
    "status",
]


class LLMStaticExporter:
    """Handle exports of LLM_RLHF projects."""

    def __init__(self, kili_api_gateway: KiliAPIGateway, disable_tqdm: Optional[bool]):
        self.kili_api_gateway = kili_api_gateway
        self.disable_tqdm = disable_tqdm

    def export(
        self, project_id: ProjectId, asset_filter: AssetFilters, json_interface: Dict
    ) -> List[Dict[str, Union[List[str], str]]]:
        """Assets are static, with n labels."""
        assets = list(
            self.kili_api_gateway.list_assets(
                asset_filter, ASSET_NEEDED_FIELDS, QueryOptions(disable_tqdm=self.disable_tqdm)
            )
        )
        with TemporaryDirectory() as tmpdirname:
            assets = MediaDownloader(
                tmpdirname,
                project_id,
                False,
                "LLM_RLHF",
                self.kili_api_gateway.http_client,
            ).download_assets(assets)
            return self._process_llm_v1(assets, json_interface)

    def _process_llm_v1(
        self, assets: List[Dict], json_interface: Dict[str, Dict]
    ) -> List[Dict[str, Union[List[str], str]]]:
        result = []
        for asset in assets:
            result.append(
                {
                    "raw_data": _format_raw_data(asset),
                    "status": asset["status"],
                    "external_id": asset["externalId"],
                    "metadata": asset["jsonMetadata"],
                    "labels": list(
                        map(
                            lambda label: {
                                "author": label["author"]["email"],
                                "created_at": label["createdAt"],
                                "label_type": label["labelType"],
                                "label": _format_json_response(
                                    json_interface["jobs"], label["jsonResponse"]
                                ),
                            },
                            asset["labels"],
                        )
                    ),
                }
            )
        return result


def _format_json_response(
    jobs_config: Dict, json_response: Dict
) -> Dict[str, Union[str, List[str]]]:
    result = {}
    for job_name, job_value in json_response.items():
        job_config = jobs_config[job_name]
        if job_config is None:
            continue
        if job_config["mlTask"] == "CLASSIFICATION":
            result[job_name] = []
            for category in job_value["categories"]:
                result[job_name].append(category["name"])
                if "children" in category:
                    for child_name, child_value in _format_json_response(
                        jobs_config, category["children"]
                    ).items():
                        result[f"{job_name}.{category['name']}.{child_name}"] = child_value
        elif job_config["mlTask"] == "TRANSCRIPTION":
            result[job_name] = job_value["text"]
        else:
            logging.warning(f"Job {job_name} with mlTask {job_config['mlTask']} not supported")
    return result


def _format_raw_data(
    asset, step_number: Optional[int] = None, all_model_keys: Optional[bool] = False
) -> List[ExportLLMItem]:
    raw_data = []

    chat_id = asset["jsonMetadata"].get("chat_id", None)

    if (
        "chat_item_ids" in asset["jsonMetadata"]
        and isinstance(asset["jsonMetadata"]["chat_item_ids"], str)
        and len(asset["jsonMetadata"]["chat_item_ids"]) > 0
    ):
        chat_items_ids = str.split(asset["jsonMetadata"]["chat_item_ids"], SEPARATOR)
        if step_number is not None:
            chat_items_ids = chat_items_ids[step_number * 3 :]
    else:
        chat_items_ids = []

    if (
        "models" in asset["jsonMetadata"]
        and isinstance(asset["jsonMetadata"]["models"], str)
        and len(asset["jsonMetadata"]["models"]) > 0
    ):
        models = str.split(asset["jsonMetadata"]["models"], SEPARATOR)
    else:
        models = []
    with Path(asset["content"]).open(encoding="utf8") as file:
        data = json.load(file)
    version = data.get("version", None)
    if version == "0.1":
        prompts = data["prompts"]
        if step_number is not None:
            prompts = [prompts[step_number]]
        for index, prompt in enumerate(prompts):
            raw_data.append(
                ExportLLMItem(
                    {
                        "role": prompt.get("title", "user"),
                        "content": prompt["prompt"],
                        "id": _safe_pop(chat_items_ids),
                        "chat_id": chat_id,
                        "model": None,
                    }
                )
            )
            raw_data.extend(
                ExportLLMItem(
                    {
                        "role": completion.get("title", "assistant"),
                        "content": completion["content"],
                        "id": _safe_pop(chat_items_ids),
                        "chat_id": chat_id,
                        "model": models[index_completion]
                        if (
                            (index == len(prompts) - 1 or all_model_keys)
                            and len(models) > index_completion
                        )
                        else None,
                    }
                )
                for index_completion, completion in enumerate(prompt["completions"])
            )
    else:
        raise ValueError(f"Version {version} not supported")
    return raw_data


def _safe_pop(lst, index=0):
    try:
        return lst.pop(index)
    except IndexError:
        return None



"""Mixin extending Kili API Gateway class with Issue related operations."""

from typing import Any, Dict, Generator, List

from kili.adapters.kili_api_gateway.base import BaseOperationMixin
from kili.adapters.kili_api_gateway.helpers.queries import (
    PaginatedGraphQLQuery,
    QueryOptions,
    fragment_builder,
)
from kili.adapters.kili_api_gateway.issue.operations import (
    GQL_COUNT_ISSUES,
    GQL_CREATE_ISSUES,
)
from kili.adapters.kili_api_gateway.issue.types import IssueToCreateKiliAPIGatewayInput
from kili.core.constants import MUTATION_BATCH_SIZE
from kili.core.utils.pagination import batcher
from kili.domain.issue import IssueFilters, IssueId, IssueStatus, IssueType
from kili.domain.types import ListOrTuple
from kili.utils import tqdm

from .mappers import issue_where_mapper
from .operations import GQL_UPDATE_ISSUE, get_issues_query


class IssueOperationMixin(BaseOperationMixin):
    """GraphQL Mixin extending GraphQL Gateway class with Issue related operations."""

    def create_issues(
        self, type_: IssueType, issues: List[IssueToCreateKiliAPIGatewayInput], description: str
    ) -> List[IssueId]:
        """Send a GraphQL request calling createIssues resolver."""
        created_issue_entities: List[IssueId] = []
        with tqdm.tqdm(total=len(issues), desc=description) as pbar:
            for issues_batch in batcher(issues, batch_size=MUTATION_BATCH_SIZE):
                batch_targeted_asset_ids = [issue.asset_id for issue in issues_batch]
                payload = {
                    "issues": [
                        {
                            "issueNumber": 0,
                            "labelID": issue.label_id,
                            "objectMid": issue.object_mid,
                            "type": type_,
                            "assetId": issue.asset_id,
                            "text": issue.text,
                        }
                        for issue in issues_batch
                    ],
                    "where": {"idIn": batch_targeted_asset_ids},
                }
                result = self.graphql_client.execute(GQL_CREATE_ISSUES, payload)
                batch_created_issues = result["data"]
                created_issue_entities.extend(
                    [IssueId(issue["id"]) for issue in batch_created_issues]
                )
                pbar.update(len(issues_batch))
        return created_issue_entities

    def count_issues(self, filters: IssueFilters) -> int:
        """Send a GraphQL request calling countIssues resolver."""
        where = issue_where_mapper(filters)
        payload = {"where": where}
        count_result = self.graphql_client.execute(GQL_COUNT_ISSUES, payload)
        return count_result["data"]

    def list_issues(
        self, filters: IssueFilters, fields: ListOrTuple[str], options: QueryOptions
    ) -> Generator[Dict, None, None]:
        """Send a GraphQL request calling issues resolver."""
        fragment = fragment_builder(fields)
        query = get_issues_query(fragment)
        where = issue_where_mapper(filters=filters)
        return PaginatedGraphQLQuery(self.graphql_client).execute_query_from_paginated_call(
            query, where, options, "Retrieving issues", GQL_COUNT_ISSUES
        )

    def update_issue_status(self, issue_id: IssueId, status: IssueStatus) -> Dict[str, Any]:
        """Update the status of an issue."""
        data = {"status": status}
        where = {"id": issue_id}
        payload = {"data": data, "where": where}
        return self.graphql_client.execute(GQL_UPDATE_ISSUE, payload)



"""Formatters for labels retrieved from Kili API."""

import json
from typing import Dict

from kili.domain.types import ListOrTuple


def load_label_json_fields(label: Dict, fields: ListOrTuple[str]) -> Dict:
    """Load json fields of a label."""
    if "jsonResponse" in fields:
        try:
            label["jsonResponse"] = json.loads(label.get("jsonResponse", "{}"))
        except json.JSONDecodeError:
            label["jsonResponse"] = {}

    return label



"""Issue use cases."""

from typing import Any, Dict, Generator, List

from kili.adapters.kili_api_gateway.helpers.queries import QueryOptions
from kili.adapters.kili_api_gateway.issue.types import IssueToCreateKiliAPIGatewayInput
from kili.domain.issue import IssueFilters, IssueId, IssueStatus
from kili.domain.project import ProjectId
from kili.domain.types import ListOrTuple
from kili.entrypoints.mutations.issue.helpers import get_labels_asset_ids_map
from kili.use_cases.base import BaseUseCases
from kili.use_cases.issue.types import IssueToCreateUseCaseInput


class IssueUseCases(BaseUseCases):
    """Issue use cases."""

    def create_issues(
        self, project_id: ProjectId, issues: List[IssueToCreateUseCaseInput]
    ) -> List[IssueId]:
        """Create issues with issue type."""
        label_id_array = [issue.label_id for issue in issues]
        label_asset_ids_map = get_labels_asset_ids_map(
            self._kili_api_gateway, project_id, label_id_array
        )  # TODO: should be done in the backend
        gateway_issues = [
            IssueToCreateKiliAPIGatewayInput(
                label_id=issue.label_id,
                object_mid=issue.object_mid,
                asset_id=label_asset_ids_map[issue.label_id],
                text=issue.text,
            )
            for issue in issues
        ]
        return self._kili_api_gateway.create_issues(
            type_="ISSUE", issues=gateway_issues, description="Creating issues"
        )

    def count_issues(self, filters: IssueFilters) -> int:
        """Count issues."""
        return self._kili_api_gateway.count_issues(filters)

    def list_issues(
        self, filters: IssueFilters, fields: ListOrTuple[str], options: QueryOptions
    ) -> Generator[Dict, None, None]:
        """List issues."""
        return self._kili_api_gateway.list_issues(filters=filters, fields=fields, options=options)

    def update_issue_status(self, issue_id: IssueId, status: IssueStatus) -> Dict[str, Any]:
        """Update issue status."""
        return self._kili_api_gateway.update_issue_status(issue_id=issue_id, status=status)



"""Custom Types."""

from enum import Enum
from typing import List, Optional, TypedDict


class ExportLLMItem(TypedDict):
    """LLM asset chat part."""

    role: str
    content: str
    id: Optional[str]
    chat_id: Optional[str]
    model: Optional[str]


class ExportLLMAsset(TypedDict):
    """LLM export asset format."""

    raw_data: List[ExportLLMItem]


class RankingValue(str, Enum):
    """Possible value for ranking."""

    A_3 = "A_3"
    A_2 = "A_2"
    A_1 = "A_1"
    TIE = "TIE"
    B_1 = "B_1"
    B_2 = "B_2"
    B_3 = "B_3"



"""Project domain."""

from dataclasses import dataclass
from enum import Enum
from typing import TYPE_CHECKING, Literal, NewType, Optional

from .types import ListOrTuple

if TYPE_CHECKING:
    from .tag import TagId

ProjectId = NewType("ProjectId", str)
InputType = Literal["IMAGE", "PDF", "TEXT", "VIDEO", "LLM_RLHF", "LLM_INSTR_FOLLOWING"]


class InputTypeEnum(str, Enum):
    """Input type enum."""

    IMAGE = "IMAGE"
    PDF = "PDF"
    TEXT = "TEXT"
    VIDEO = "VIDEO"
    LLM_RLHF = "LLM_RLHF"
    LLM_INSTR_FOLLOWING = "LLM_INSTR_FOLLOWING"


ComplianceTag = Literal["PHI", "PII"]


@dataclass
# pylint: disable=too-many-instance-attributes
class ProjectFilters:
    """Project filters for running a project search."""

    id: Optional[ProjectId]
    archived: Optional[bool] = None
    search_query: Optional[str] = None
    should_relaunch_kpi_computation: Optional[bool] = None
    starred: Optional[bool] = None
    updated_at_gte: Optional[str] = None
    updated_at_lte: Optional[str] = None
    created_at_gte: Optional[str] = None
    created_at_lte: Optional[str] = None
    organization_id: Optional[str] = None
    tag_ids: Optional[ListOrTuple["TagId"]] = None



"""CLI's Project Member common functions."""

import re
import warnings
from typing import TYPE_CHECKING, Callable, Iterable, Optional

from kili.adapters.kili_api_gateway.helpers.queries import QueryOptions
from kili.client import Kili
from kili.core.graphql.operations.project_user.queries import (
    ProjectUserQuery,
    ProjectUserWhere,
)
from kili.entrypoints.cli.common_args import ROLES
from kili.entrypoints.cli.helpers import collect_from_csv

if TYPE_CHECKING:
    from kili.adapters.http_client import HttpClient

REGEX_EMAIL = re.compile(r"([A-Za-z0-9]+[.-_])*[A-Za-z0-9]+@[A-Za-z0-9-]+(\.[A-Z|a-z]{2,})+")


def type_check_member(key, value):
    """Type check value based on key."""
    if key == "email" and not re.search(REGEX_EMAIL, value):
        return f"{value} is not a valid email address, "

    if key == "role" and value not in ROLES:
        return f"{value} is not a valid role, "

    return ""


def collect_members_from_csv(csv_path: str, role: Optional[str]):
    """Read a csv with to collect members and role."""
    type_check_function: Callable[[str, str, Optional[HttpClient]], str] = (
        lambda key, value, z: type_check_member(key, value)
    )
    members_to_add = collect_from_csv(
        csv_path=csv_path,
        required_columns=["email"],
        optional_columns=["role"],
        type_check_function=type_check_function,
    )

    if len(members_to_add) == 0:
        raise ValueError(f"No active member were found in csv: {csv_path}")

    if "role" in members_to_add[0]:
        if role is not None:
            raise ValueError(
                "--role cannot be used if the argument passed is a path to a csv file with roles"
            )
    else:
        if role is None:
            role = "LABELER"
        for member in members_to_add:
            member["role"] = role

    return members_to_add


def collect_members_from_project(kili: Kili, project_id_source: str, role: Optional[str]):
    """Copy members from project of id project_id_source."""
    activated_members = []

    if role is not None:
        raise ValueError("--role cannot be used if the argument passed is a Kili project_id")

    existing_members = ProjectUserQuery(kili.graphql_client, kili.http_client)(
        where=ProjectUserWhere(project_id=project_id_source),
        fields=["role", "user.email", "activated"],
        options=QueryOptions(disable_tqdm=True),
    )
    for existing_member in existing_members:
        if existing_member["activated"]:
            activated_members.append(
                {"email": existing_member["user"]["email"], "role": existing_member["role"]}
            )

    if len(activated_members) == 0:
        raise ValueError(
            f"No active member were found in project with id {project_id_source} or the project"
            " does not exist"
        )

    return activated_members


def collect_members_from_emails(emails: Iterable[str], role: Optional[str]):
    """Collect members with email address from emails."""
    if role is None:
        role = "LABELER"
    members_to_add = []
    for email in emails:
        if re.search(REGEX_EMAIL, email):
            members_to_add.append({"email": email, "role": role})
        else:
            warnings.warn(f"{email} is not a valid email address,", stacklevel=2)

    if len(members_to_add) == 0:
        raise ValueError("No valid email adresses were provided")

    return members_to_add


def check_exclusive_options(
    csv_path: Optional[str],
    project_id_src: Optional[str],
    emails: Optional[Iterable[str]],
    all_members: Optional[bool],
) -> None:
    """Forbid mutual use of options and argument(s)."""
    if not emails:
        return
    emails = list(emails)
    if all_members is None:
        if (csv_path is not None) + (project_id_src is not None) + (len(emails) > 0) > 1:
            raise ValueError(
                "emails arguments and options --from-csv and --from-project are exclusive."
            )
        if (csv_path is not None) + (project_id_src is not None) + (len(emails) > 0) == 0:
            raise ValueError(
                "You must either provide emails arguments or use one option from "
                "--from-csv or --from-project"
            )

    else:
        if (csv_path is not None) + (all_members is True) + (len(emails) > 0) > 1:
            raise ValueError("emails arguments and options --from-csv and --all are exclusive.")
        if (csv_path is not None) + (all_members is not None) + (len(emails) > 0) == 0:
            raise ValueError(
                "You must either provide emails arguments or use one option --from-csv or --all"
            )
    return



"""Module for managing bucket's signed urls."""

import itertools
from typing import List, Union
from urllib.parse import parse_qs, urlparse

import cuid
from tenacity import retry
from tenacity.stop import stop_after_attempt
from tenacity.wait import wait_random

from kili.adapters.http_client import HttpClient

MAX_NUMBER_SIGNED_URLS_TO_FETCH = 30


def generate_unique_id() -> str:
    """Generate a unique id."""
    return cuid.cuid()


# pylint: disable=missing-type-doc
def request_signed_urls(kili, file_urls: List[str]) -> List[str]:
    """Get upload signed URLs.

    Args:
        kili: Kili
        file_urls: the paths in Kili bucket of the data you upload. It must respect
            the convention projects/:projectId/assets/xxx.
    """
    size = len(file_urls)
    file_batches = [
        file_urls[i : i + MAX_NUMBER_SIGNED_URLS_TO_FETCH]
        for i in range(0, size, MAX_NUMBER_SIGNED_URLS_TO_FETCH)
    ]

    request_function = kili.kili_api_gateway.create_upload_bucket_signed_urls

    return [*itertools.chain(*map(request_function, file_batches))]


@retry(stop=stop_after_attempt(3), wait=wait_random(min=1, max=2), reraise=True)
def upload_data_via_rest(
    url_with_id: str, data: Union[str, bytes], content_type: str, http_client: HttpClient
) -> str:
    """Upload data in buckets' signed URL via REST.

    Args:
        url_with_id: signed url with id
        data: data to upload
        content_type: mimetype of the data
        http_client: http client
    """
    if content_type == "text/plain":
        content_type += "; charset=utf-8"
    headers = {"Content-type": content_type}
    url_to_use_for_upload = url_with_id.split("&id=")[0]
    if "blob.core.windows.net" in url_to_use_for_upload:
        headers["x-ms-blob-type"] = "BlockBlob"
    # Do we not put a timeout here because it can take an arbitrary long time (ML-1395)
    response = http_client.put(url_to_use_for_upload, data=data, headers=headers)
    response.raise_for_status()
    return url_with_id


def clean_signed_url(url: str, endpoint: str) -> str:
    """Return a cleaned signed url for frame upload."""
    query = urlparse(url).query
    id_param = parse_qs(query)["id"][0]
    base_path = endpoint.replace("/graphql", "/files").replace("http://", "https://")
    return f"{base_path}?id={id_param}"



"""Base class for all formatters and other utility classes."""

import csv
import json
import logging
import shutil
from abc import ABC, abstractmethod
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING, Dict, List, NamedTuple, Optional, Tuple, Union

from kili.domain.asset import AssetId
from kili.domain.project import ProjectId
from kili.services.export.exceptions import (
    NotCompatibleOptions,
)
from kili.services.export.repository import AbstractContentRepository
from kili.services.export.tools import (
    fetch_assets,
    is_geotiff_asset_with_lat_lon_coords,
)
from kili.services.export.types import (
    CocoAnnotationModifier,
    ExportType,
    LabelFormat,
    SplitOption,
)
from kili.services.types import Job
from kili.utils.tempfile import TemporaryDirectory

if TYPE_CHECKING:
    from kili.client import Kili


class ExportParams(NamedTuple):
    """Contains all parameters that change the result of the export."""

    assets_ids: Optional[List[AssetId]]
    export_type: ExportType
    project_id: ProjectId
    label_format: LabelFormat
    split_option: SplitOption
    single_file: bool
    output_file: Optional[Path]
    with_assets: bool
    annotation_modifier: Optional[CocoAnnotationModifier]
    asset_filter_kwargs: Optional[Dict[str, object]]
    normalized_coordinates: Optional[bool]
    label_type_in: Optional[List[str]]


class AbstractExporter(ABC):  # pylint: disable=too-many-instance-attributes
    """Abstract class defining the interface for all exporters."""

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        export_params: ExportParams,
        kili: "Kili",
        logger: logging.Logger,
        disable_tqdm: Optional[bool],
        content_repository: AbstractContentRepository,
    ) -> None:
        """Initialize the exporter."""
        self.project_id: ProjectId = export_params.project_id
        self.assets_ids: Optional[List[AssetId]] = export_params.assets_ids
        self.export_type: ExportType = export_params.export_type
        self.label_format: LabelFormat = export_params.label_format
        self.single_file: bool = export_params.single_file
        self.split_option: SplitOption = export_params.split_option
        self.disable_tqdm: Optional[bool] = disable_tqdm
        self.kili = kili
        self.logger: logging.Logger = logger
        self.content_repository: AbstractContentRepository = content_repository
        self.output_file = export_params.output_file
        self.with_assets: bool = export_params.with_assets
        self.export_root_folder: Path = Path()
        self.annotation_modifier = export_params.annotation_modifier
        self.asset_filter_kwargs = export_params.asset_filter_kwargs
        self.normalized_coordinates = export_params.normalized_coordinates
        self.label_type_in = export_params.label_type_in or ["DEFAULT", "REVIEW"]

        self.project = kili.kili_api_gateway.get_project(
            self.project_id, ["jsonInterface", "inputType", "title", "description", "id"]
        )

    @abstractmethod
    def _check_arguments_compatibility(self) -> None:
        """Checks if the export label format is compatible with the export options."""

    @abstractmethod
    def _check_project_compatibility(self) -> None:
        """Checks if the export label format is compatible with the project."""

    @abstractmethod
    def _is_job_compatible(self, job: Job) -> bool:
        """Check if the export label format is compatible with the job."""

    @property
    def compatible_jobs(self) -> Tuple[str, ...]:
        """Get all job names compatible with the export format."""
        return tuple(
            job_name
            for job_name, job in self.project["jsonInterface"]["jobs"].items()
            if self._is_job_compatible(job)
        )

    @abstractmethod
    def process_and_save(self, assets: List[Dict], output_filename: Path) -> None:
        """Converts the asset and save them into an archive file."""

    def process(self, assets: List[Dict]) -> List[Dict[str, Union[List[str], str]]]:
        """Converts the asset."""
        raise ValueError("Output file is required for this export format.")

    def make_archive(self, root_folder: Path, output_filename: Path) -> Path:
        """Make the export archive."""
        path_folder = root_folder / self.project_id
        path_archive = shutil.make_archive(str(path_folder), "zip", path_folder)
        output_filename.parent.mkdir(parents=True, exist_ok=True)
        shutil.copy(path_archive, output_filename)
        return output_filename

    def create_readme_kili_file(self, root_folder: Path) -> None:
        """Create a README.kili.txt file to give information about exported labels."""
        readme_file_name = root_folder / self.project_id / "README.kili.txt"
        readme_file_name.parent.mkdir(parents=True, exist_ok=True)
        with readme_file_name.open("wb") as fout:
            fout.write(b"Exported Labels from KILI\n=========================\n\n")
            fout.write(f"- Project name: {self.project['title']}\n".encode())
            fout.write(f"- Project identifier: {self.project['id']}\n".encode())
            fout.write(f"- Project description: {self.project.get('description', '')}\n".encode())
            fout.write(f'- Export date: {datetime.now().strftime(r"%Y%m%d-%H%M%S")}\n'.encode())
            fout.write(f"- Exported format: {self.label_format}\n".encode())
            fout.write(f"- Exported labels: {self.export_type}\n".encode())

    @staticmethod
    def write_video_metadata_file(video_metadata: Dict, base_folder: Path) -> None:
        """Write video metadata file."""
        video_metadata_json = json.dumps(video_metadata, sort_keys=True, indent=4)
        if video_metadata_json is not None:
            with (base_folder / "video_meta.json").open("wb") as output_file:
                output_file.write(video_metadata_json.encode("utf-8"))

    @staticmethod
    def write_remote_content_file(remote_content: List[str], images_folder: Path) -> None:
        """Write remote content file."""
        remote_content_header = ["external id", "url", "label file"]
        # newline="" to disable universal newlines translation (bug fix for windows)
        with (images_folder / "remote_assets.csv").open("w", newline="", encoding="utf8") as file:
            writer = csv.writer(file)
            writer.writerow(remote_content_header)
            writer.writerows(remote_content)

    def export_project(
        self,
    ) -> Optional[List[Dict[str, Union[List[str], str]]]]:
        """Export a project to a json.

        Return the name of the exported archive file.
        """
        self._check_arguments_compatibility()
        self._check_project_compatibility()
        self._check_and_ensure_asset_access()

        self.logger.info("Fetching assets...")

        with TemporaryDirectory() as export_root_folder:
            self.export_root_folder = export_root_folder
            assets = fetch_assets(
                self.kili,
                project_id=self.project_id,
                asset_ids=self.assets_ids,
                export_type=self.export_type,
                label_type_in=self.label_type_in,
                disable_tqdm=self.disable_tqdm,
                download_media=self.with_assets,
                local_media_dir=str(self.images_folder),
                asset_filter_kwargs=self.asset_filter_kwargs,
            )

            self._check_geotiff_export_compatibility(assets)

            if self.output_file is None:
                return self.process(assets)

            return self.process_and_save(assets, self.output_file)

    def _check_and_ensure_asset_access(self) -> None:
        """Check asset access.

        If there is a data connection, and that the format requires a data access, or that
        with assets is passed, then output an error.

        If not, if the format requires a data access, ensure that the assets are requested.
        """
        if self._has_data_connection() and self.with_assets:
            raise NotCompatibleOptions(
                "Export with download of assets is not allowed on projects with data connections."
                " Please disable the download of assets by setting `with_assets=False`."
            )

    def _has_data_connection(self) -> bool:
        project = self.kili.kili_api_gateway.get_project(self.project_id, ["dataConnections.id"])
        return bool(project["dataConnections"])

    def _check_geotiff_export_compatibility(self, assets: List[Dict]) -> None:
        # pylint: disable=line-too-long
        """Check if one of the assets is a geotiff asset, and if the export params are compatible.

        If one of the assets is a geotiff asset, then:

        - we can only allow the export for Kili or geojson formats, since geotiff normalizedVertices are lat/lon coordinates
        - we cannot allow normalized_coordinates != None, for the same reason
        """
        if (
            self.label_format not in ("raw", "kili", "geojson")
            or self.normalized_coordinates is not None
        ):
            has_geotiff_asset = any(
                is_geotiff_asset_with_lat_lon_coords(asset, self.kili.http_client)
                for asset in assets
            )

            if self.label_format not in ("raw", "kili", "geojson") and has_geotiff_asset:
                raise NotCompatibleOptions(
                    "Cannot export geotiff assets with geospatial coordinates in"
                    f" {self.label_format} format. Please use 'kili', 'raw' or 'geojson' formats"
                    " instead."
                )

            if self.normalized_coordinates is not None and has_geotiff_asset:
                raise NotCompatibleOptions(
                    "Cannot export geotiff assets with geospatial latitude and longitude"
                    f" coordinates with normalized_coordinates={self.normalized_coordinates}."
                    " Please use `normalized_coordinates=None` instead."
                )

    @property
    def base_folder(self) -> Path:
        """Export base folder."""
        return self.export_root_folder / self.project_id

    @property
    def images_folder(self) -> Path:
        """Export images folder."""
        return self.base_folder / "images"

    @staticmethod
    def _filter_out_autosave_labels(assets: List[Dict]) -> List[Dict]:
        """Removes AUTOSAVE labels from exports."""
        clean_assets = []
        for asset in assets:
            labels = asset.get("labels", [])
            clean_labels = list(filter(lambda label: label["labelType"] != "AUTOSAVE", labels))
            if clean_labels:
                asset["labels"] = clean_labels
            clean_assets.append(asset)
        return clean_assets

    @staticmethod
    def _format_json_response(label: Dict) -> Dict:
        """Format the label JSON response in the requested format."""
        formatted_json_response = label["jsonResponse"]
        json_response = {}
        for key, value in formatted_json_response.items():
            if key.isdigit():
                json_response[int(key)] = value
                continue
            json_response[key] = value
        label["jsonResponse"] = json_response
        return label

    def preprocess_assets(self, assets: List[Dict]) -> List[Dict]:
        """Format labels in the requested format, and filter out autosave labels."""
        assets_in_format = []
        for asset in assets:
            if "labels" in asset:
                labels_of_asset = []
                for label in asset["labels"]:
                    clean_label = AbstractExporter._format_json_response(label)
                    labels_of_asset.append(clean_label)
                asset["labels"] = labels_of_asset
            if "latestLabel" in asset:
                label = asset["latestLabel"]
                if label is not None:
                    clean_label = AbstractExporter._format_json_response(label)
                    asset["latestLabel"] = clean_label
            assets_in_format.append(asset)

        return AbstractExporter._filter_out_autosave_labels(assets_in_format)



"""Validator for import of labels."""

from typing import List

from typeguard import typechecked

from .types import LabelToCreateUseCaseInput


@typechecked
def check_input_labels(labels: List[LabelToCreateUseCaseInput]) -> None:
    """Check that input labels are valid."""
    for label in labels:
        _check_input_label(label)

    is_using_asset_id = labels[0].asset_id is not None
    if is_using_asset_id and not all(label.asset_id for label in labels):
        raise ValueError(
            "Please use the same asset identifier for all labels: either only `asset_id` or"
            " only `asset_external_id`."
        )


@typechecked
def _check_input_label(label: LabelToCreateUseCaseInput) -> None:
    """Check that a single input label is valid."""
    if label.asset_external_id is None and label.asset_id is None:
        raise ValueError("You must either provide the `asset_id` or the `external_id`.")

    if label.label_type == "PREDICTION" and not label.model_name:
        raise ValueError("You must provide `model_name` when uploading `PREDICTION` labels.")



"""Types for label-related use cases."""

from dataclasses import dataclass
from typing import Dict, Optional

from kili.domain.asset import AssetExternalId, AssetId
from kili.domain.label import LabelType
from kili.domain.user import UserId


@dataclass
class LabelToCreateUseCaseInput:
    """Data about one label to create."""

    asset_id: Optional[AssetId]
    asset_external_id: Optional[AssetExternalId]
    label_type: LabelType
    json_response: Dict
    author_id: Optional[UserId]
    seconds_to_label: Optional[float]
    model_name: Optional[str]



"""Mixin extending Kili API Gateway class with Api Keys related operations."""

from typing import Dict, List, Optional, cast

from kili.adapters.kili_api_gateway.base import BaseOperationMixin
from kili.adapters.kili_api_gateway.helpers.queries import (
    PaginatedGraphQLQuery,
    QueryOptions,
    fragment_builder,
)
from kili.adapters.kili_api_gateway.llm.mappers import (
    map_asset_where,
    map_create_chat_item_input,
    map_create_llm_asset_input,
    map_create_model_input,
    map_create_project_model_input,
    map_delete_model_input,
    map_delete_project_model_input,
    map_project_where,
    map_update_model_input,
    map_update_project_model_input,
    model_where_wrapper,
    project_model_where_mapper,
)
from kili.adapters.kili_api_gateway.llm.operations import (
    get_create_chat_item_mutation,
    get_create_llm_asset_mutation,
    get_create_model_mutation,
    get_create_project_model_mutation,
    get_delete_model_mutation,
    get_delete_project_model_mutation,
    get_model_query,
    get_models_query,
    get_project_models_query,
    get_update_model_mutation,
    get_update_project_model_mutation,
)
from kili.domain.llm import (
    ChatItemDict,
    ModelDict,
    ModelToCreateInput,
    ModelToUpdateInput,
    OrganizationModelFilters,
    ProjectModelDict,
    ProjectModelFilters,
    ProjectModelToCreateInput,
    ProjectModelToUpdateInput,
)
from kili.domain.types import ListOrTuple

DEFAULT_PROJECT_FIELDS = ["id", "name", "credentials", "type"]
DEFAULT_PROJECT_MODEL_FIELDS = [
    "id",
    "configuration",
    "model.id",
    "model.credentials",
    "model.name",
    "model.type",
]


class ModelConfigurationOperationMixin(BaseOperationMixin):
    """Mixin extending Kili API Gateway class with model configuration related operations."""

    def list_models(
        self,
        filters: OrganizationModelFilters,
        fields: Optional[ListOrTuple[str]] = None,
        options: Optional[QueryOptions] = None,
    ) -> List[ModelDict]:
        """List models with given options."""
        fragment = fragment_builder(fields or DEFAULT_PROJECT_FIELDS)
        query = get_models_query(fragment)
        where = model_where_wrapper(filters)
        return [
            cast(ModelDict, item)
            for item in PaginatedGraphQLQuery(
                self.graphql_client
            ).execute_query_from_paginated_call(
                query,
                where,
                options if options else QueryOptions(disable_tqdm=False),
                "Retrieving organization models",
                None,
            )
        ]

    def get_model(self, model_id: str, fields: Optional[ListOrTuple[str]] = None) -> ModelDict:
        """Get a model by ID."""
        fragment = fragment_builder(fields or DEFAULT_PROJECT_FIELDS)
        query = get_model_query(fragment)
        variables = {"modelId": model_id}
        result = self.graphql_client.execute(query, variables)
        return result["model"]

    def create_model(self, model: ModelToCreateInput) -> ModelDict:
        """Send a GraphQL request calling createModel resolver."""
        payload = {"input": map_create_model_input(model)}
        fragment = fragment_builder(DEFAULT_PROJECT_FIELDS)
        mutation = get_create_model_mutation(fragment)
        result = self.graphql_client.execute(mutation, payload)
        return result["createModel"]

    def update_properties_in_model(self, model_id: str, model: ModelToUpdateInput) -> ModelDict:
        """Send a GraphQL request calling updateModel resolver."""
        payload = {"id": model_id, "input": map_update_model_input(model)}
        fragment = fragment_builder(DEFAULT_PROJECT_FIELDS)
        mutation = get_update_model_mutation(fragment)
        result = self.graphql_client.execute(mutation, payload)
        return result["updateModel"]

    def delete_model(self, model_id: str) -> bool:
        """Send a GraphQL request to delete an organization model."""
        payload = map_delete_model_input(model_id)
        mutation = get_delete_model_mutation()
        result = self.graphql_client.execute(mutation, payload)
        return result["deleteModel"]

    def create_project_model(self, project_model: ProjectModelToCreateInput) -> ProjectModelDict:
        """Send a GraphQL request calling createModel resolver."""
        payload = {"input": map_create_project_model_input(project_model)}
        fragment = fragment_builder(DEFAULT_PROJECT_MODEL_FIELDS)
        mutation = get_create_project_model_mutation(fragment)
        result = self.graphql_client.execute(mutation, payload)
        return result["createProjectModel"]

    def update_project_model(
        self, project_model_id: str, project_model: ProjectModelToUpdateInput
    ) -> ProjectModelDict:
        """Send a GraphQL request calling updateProjectModel resolver."""
        payload = {
            "updateProjectModelId": project_model_id,
            "input": map_update_project_model_input(project_model),
        }
        fragment = fragment_builder(DEFAULT_PROJECT_MODEL_FIELDS)
        mutation = get_update_project_model_mutation(fragment)
        result = self.graphql_client.execute(mutation, payload)
        return result["updateProjectModel"]

    def delete_project_model(self, project_model_id: str) -> bool:
        """Send a GraphQL request to delete a project model."""
        payload = map_delete_project_model_input(project_model_id)
        mutation = get_delete_project_model_mutation()
        result = self.graphql_client.execute(mutation, payload)
        return result["deleteProjectModel"]

    def list_project_models(
        self,
        filters: ProjectModelFilters,
        fields: Optional[ListOrTuple[str]] = None,
        options: Optional[QueryOptions] = None,
    ) -> List[ProjectModelDict]:
        """List project models with given options."""
        fragment = fragment_builder(fields or DEFAULT_PROJECT_MODEL_FIELDS)
        query = get_project_models_query(fragment)
        where = project_model_where_mapper(filters)
        return [
            cast(ProjectModelDict, item)
            for item in PaginatedGraphQLQuery(
                self.graphql_client
            ).execute_query_from_paginated_call(
                query,
                where,
                options if options else QueryOptions(disable_tqdm=False),
                "Retrieving project models",
                None,
            )
        ]

    def create_llm_asset(
        self,
        project_id: str,
        author_id: str,
        status: Optional[str] = None,
        label_type: Optional[str] = None,
    ) -> Dict:
        """Create an LLM asset in a project, with optional status and label_type."""
        where = map_project_where(project_id)
        data = {"author_id": author_id, "status": status, "label_type": label_type}
        data_mapped = map_create_llm_asset_input(data)
        variables = {"where": where, "data": data_mapped}
        fragment = fragment_builder(["id", "latestLabel.id"])
        mutation = get_create_llm_asset_mutation(fragment)
        result = self.graphql_client.execute(mutation, variables)
        return result["createLLMAsset"]

    def create_chat_item(self, asset_id: str, label_id: str, prompt: str) -> List[ChatItemDict]:
        """Create a chat item associated with an asset."""
        data = map_create_chat_item_input(label_id, prompt)
        where = map_asset_where(asset_id)
        variables = {"data": data, "where": where}
        fragment = fragment_builder(["content", "id", "labelId", "modelId", "parentId", "role"])
        mutation = get_create_chat_item_mutation(fragment)
        result = self.graphql_client.execute(mutation, variables)
        return [cast(ChatItemDict, item) for item in result["createChatItem"]]



"""Plugins queries."""

import json
from datetime import datetime
from typing import Dict, List, Optional

from typeguard import typechecked

from kili.adapters.kili_api_gateway.helpers.queries import QueryOptions
from kili.core.graphql.operations.plugin.queries import (
    PluginBuildErrorsWhere,
    PluginLogsWhere,
    PluginQuery,
)
from kili.domain.types import ListOrTuple
from kili.entrypoints.base import BaseOperationEntrypointMixin
from kili.services.plugins import PluginUploader
from kili.utils.logcontext import for_all_methods, log_call


@for_all_methods(log_call, exclude=["__init__"])
class QueriesPlugins(BaseOperationEntrypointMixin):
    """Set of Plugins queries."""

    # pylint: disable=too-many-arguments

    @typechecked
    def get_plugin_build_errors(
        self,
        plugin_name: str,
        start_date: Optional[datetime] = None,
        limit: int = 100,
        skip: int = 0,
    ) -> str:
        # pylint: disable=line-too-long
        """Get paginated build errors of a plugin.

        Args:
            plugin_name: Name of the plugin
            start_date: Datetime used to get the build errors from, if not provided, it will be the plugin's creation date
            limit: Limit for pagination, if not provided, it will be 100
            skip: Skip for pagination, if not provided, it will be 0
        Returns:
            A result array which contains the build errors of the plugin,
                or an error message.

        Examples:
            >>> kili.get_plugin_build_errors(plugin_name="my_plugin_name", start_date="1970/01/01")
        """
        where = PluginBuildErrorsWhere(plugin_name=plugin_name, start_date=start_date)
        options = QueryOptions(
            first=limit, skip=skip, disable_tqdm=False
        )  # disable tqm is not implemented for this query
        pretty_result = PluginQuery(self.graphql_client, self.http_client).get_build_errors(
            where, options
        )
        return json.dumps(pretty_result, sort_keys=True, indent=4)

    @typechecked
    def get_plugin_logs(
        self,
        project_id: str,
        plugin_name: str,
        start_date: Optional[datetime] = None,
        limit: int = 100,
        skip: int = 0,
    ) -> str:
        # pylint: disable=line-too-long
        """Get paginated logs of a plugin on a project.

        Args:
            project_id: Identifier of the project
            plugin_name: Name of the plugin
            start_date: Datetime used to get the logs from, if not provided, it will be the plugin's creation date
            limit: Limit for pagination, if not provided, it will be 100
            skip: Skip for pagination, if not provided, it will be 0
        Returns:
            A result array which contains the logs of the plugin,
                or an error message.

        Examples:
            >>> kili.get_plugin_logs(project_id="my_project_id", plugin_name="my_plugin_name", start_date="1970/01/01")
        """
        where = PluginLogsWhere(
            project_id=project_id, plugin_name=plugin_name, start_date=start_date
        )
        options = QueryOptions(
            first=limit, skip=skip, disable_tqdm=False
        )  # disable tqm is not implemented for this query
        pretty_result = PluginQuery(self.graphql_client, self.http_client).get_logs(where, options)
        return json.dumps(pretty_result, sort_keys=True, indent=4)

    @typechecked
    def get_plugin_status(
        self,
        plugin_name: str,
        verbose: bool = True,
    ) -> str:
        """Update a plugin with new code.

        Args:
            plugin_name: Name of the plugin
            verbose: If false, minimal logs are displayed

        Returns:
            The status of the plugin if query was successful or an error message otherwise.

        Examples:
            >>> kili.get_plugin_status(plugin_name="my_plugin_name")
        """
        return PluginUploader(
            self,  # pyright: ignore[reportGeneralTypeIssues]
            "",
            plugin_name,
            verbose,
            self.http_client,
        ).get_plugin_runner_status()

    @typechecked
    def list_plugins(
        self,
        fields: ListOrTuple[str] = ("name", "projectIds", "id", "createdAt", "updatedAt"),
    ) -> List[Dict]:
        # pylint: disable=line-too-long
        """List all plugins from your organization.

        Args:
            fields: All the fields to request among the possible fields for the plugins
                See [the documentation](https://docs.kili-technology.com/reference/graphql-api#plugins) for all possible fields.

        Returns:
            A result array which contains all the plugins from your organization,
                or an error message.

        Examples:
            >>> kili.list_plugins()
            >>> kili.list_plugins(fields=['name'])
        """
        return PluginQuery(self.graphql_client, self.http_client).list(fields=fields)
